import itertools
from typing import Iterator, List, TypeVar

from .rest import get_pagination_query_param

T = TypeVar("T")


class ListableResource(List[T]):
    pagination: Iterator[T]
    """
    Automatic paginating iterator for resources that supports it
    It handles fetching automatically the next set of data from the list

    Example:
    --------
    >>> items = Items(config).get_items()
    ... for item in items.pagination:
    ...     print_item(item)
    """

    def __init__(
        self,
        response,
        auth,
        cls=None,
        path="",
        method="GET",
        query_params={},
        max_items=10,
    ):
        has_data_response = isinstance(response, dict) and "data" in response
        is_pag = has_data_response and "links" in response

        def create(e: any):
            if cls is None:
                return e
            o = cls(**e)
            o.auth = auth
            return o

        list.__init__(
            self,
            [
                create(elem)
                for elem in (response["data"] if has_data_response else response)
            ],
        )
        if is_pag:
            pag_iter = []
            try:
                curr_params = query_params.copy()
                # should be "next" but for users it is "startingAfter"
                pagination_query_param = get_pagination_query_param(
                    url_str=response["links"]["next"]["href"]
                )
                if pagination_query_param:
                    query_key, query_value = pagination_query_param
                    curr_params[query_key] = query_value
                rem_max_items = max_items - len(self)
                pag_iter = auth.paginate_rest(
                    path=path,
                    method=method,
                    params=curr_params,
                    max_items=rem_max_items,
                )
            except:
                pass

            def gen():
                for elem in pag_iter:
                    new_elem = create(elem)
                    # add the element to the list to make it accessible after the pagination too
                    self.append(new_elem)
                    yield new_elem

            self.pagination = itertools.chain(self, gen())
        else:

            def gen():
                for e in self:
                    yield e

            self.pagination = gen()



# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import dataclass

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class GetByIDResult:
    """

    Attributes
    ----------
    contentType: str
      The type that content is encoded in, always "application/json".
    data: object
      Additional information about the event's details. The structure depends on the type and version of the event.
    eventId: str
      The event's unique identifier.
    eventTime: str
      The RFC3339 datetime when the event happened.
    eventType: str
      The type of event that describes committed action.
    eventTypeVersion: str
      The version of the event type.
    extensions: EventExtensions
      The availability of the properties depends on the event and the context it was triggered in.
    id: str
      The resource item's unique identifier.
    links: GetLinks
    source: str
      The source of the event message, usually the producing service.
    tenantId: str
      The ID of the tenant that owns the item. This is populated using the JWT.
    userId: str
      The ID of the user who performed the action that triggered the event.
    """

    contentType: str = None
    data: object = None
    eventId: str = None
    eventTime: str = None
    eventType: str = None
    eventTypeVersion: str = None
    extensions: EventExtensions = None
    id: str = None
    links: GetLinks = None
    source: str = None
    tenantId: str = None
    userId: str = None

    def __init__(self_, **kvargs):

        if "contentType" in kvargs and kvargs["contentType"] is not None:
            self_.contentType = kvargs["contentType"]
        if "data" in kvargs and kvargs["data"] is not None:
            self_.data = kvargs["data"]
        if "eventId" in kvargs and kvargs["eventId"] is not None:
            self_.eventId = kvargs["eventId"]
        if "eventTime" in kvargs and kvargs["eventTime"] is not None:
            self_.eventTime = kvargs["eventTime"]
        if "eventType" in kvargs and kvargs["eventType"] is not None:
            self_.eventType = kvargs["eventType"]
        if "eventTypeVersion" in kvargs and kvargs["eventTypeVersion"] is not None:
            self_.eventTypeVersion = kvargs["eventTypeVersion"]
        if "extensions" in kvargs and kvargs["extensions"] is not None:
            if (
                type(kvargs["extensions"]).__name__
                == GetByIDResult.__annotations__["extensions"]
            ):
                self_.extensions = kvargs["extensions"]
            else:
                self_.extensions = EventExtensions(**kvargs["extensions"])
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == GetByIDResult.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = GetLinks(**kvargs["links"])
        if "source" in kvargs and kvargs["source"] is not None:
            self_.source = kvargs["source"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "userId" in kvargs and kvargs["userId"] is not None:
            self_.userId = kvargs["userId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class EventExtensions:
    """
    The availability of the properties depends on the event and the context it was triggered in.

    Attributes
    ----------
    actor: EventExtensionsActor
      Specifies the entity performing the action on behalf of another party listed as triggering the action.
    ownerId: str
      Id of the owner of the resource affected by the eventContext.
    spaceId: str
      Id of the space related to the action performed on the eventContext.
    topLevelResourceId: str
      If the event originated from a sub resource the topLevelResourceId contains the id of the top level resource associated with the sub resource.
    updates: object
      Might be present if the action is of type "updated" and should contain information about the changes made to the resource.
    """

    actor: EventExtensionsActor = None
    ownerId: str = None
    spaceId: str = None
    topLevelResourceId: str = None
    updates: object = None

    def __init__(self_, **kvargs):

        if "actor" in kvargs and kvargs["actor"] is not None:
            if (
                type(kvargs["actor"]).__name__
                == EventExtensions.__annotations__["actor"]
            ):
                self_.actor = kvargs["actor"]
            else:
                self_.actor = EventExtensionsActor(**kvargs["actor"])
        if "ownerId" in kvargs and kvargs["ownerId"] is not None:
            self_.ownerId = kvargs["ownerId"]
        if "spaceId" in kvargs and kvargs["spaceId"] is not None:
            self_.spaceId = kvargs["spaceId"]
        if "topLevelResourceId" in kvargs and kvargs["topLevelResourceId"] is not None:
            self_.topLevelResourceId = kvargs["topLevelResourceId"]
        if "updates" in kvargs and kvargs["updates"] is not None:
            self_.updates = kvargs["updates"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class EventExtensionsActor:
    """
    Specifies the entity performing the action on behalf of another party listed as triggering the action.

    Attributes
    ----------
    sub: str
      Opaque value identifying impersonating entity.
    subType: str
      The type of the impersonating entity.
    """

    sub: str = None
    subType: str = None

    def __init__(self_, **kvargs):

        if "sub" in kvargs and kvargs["sub"] is not None:
            self_.sub = kvargs["sub"]
        if "subType" in kvargs and kvargs["subType"] is not None:
            self_.subType = kvargs["subType"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GetArchiveResult:
    """

    Attributes
    ----------
    data: list[object]
      List of archived events. The structure of the events depend on their type and version.
    """

    data: list[object] = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            self_.data = kvargs["data"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GetLinks:
    """

    Attributes
    ----------
    Self: HrefDeprecated
    self: Href
    """

    Self: HrefDeprecated = None
    self: Href = None

    def __init__(self_, **kvargs):

        if "Self" in kvargs and kvargs["Self"] is not None:
            if type(kvargs["Self"]).__name__ == GetLinks.__annotations__["Self"]:
                self_.Self = kvargs["Self"]
            else:
                self_.Self = HrefDeprecated(**kvargs["Self"])
        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == GetLinks.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = Href(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GetObjectsResult:
    """

    Attributes
    ----------
    data: list[str]
      List of requested resources.
    links: ListLinks
    """

    data: list[str] = None
    links: ListLinks = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            self_.data = kvargs["data"]
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == GetObjectsResult.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = ListLinks(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GetResult:
    """

    Attributes
    ----------
    data: list[GetByIDResult]
      List of audit items.
    links: ListLinks
    """

    data: list[GetByIDResult] = None
    links: ListLinks = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]" == GetResult.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [GetByIDResult(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == GetResult.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = ListLinks(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GetSettingsResult:
    """

    Attributes
    ----------
    data: GetSettingsResultData
      Server configuration options.
    """

    data: GetSettingsResultData = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if (
                type(kvargs["data"]).__name__
                == GetSettingsResult.__annotations__["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = GetSettingsResultData(**kvargs["data"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GetSettingsResultData:
    """
    Server configuration options.

    Attributes
    ----------
    ArchiveEnabled: bool
      Is Long Term Storage archiving enabled?.
    EventTTL: int
      The events TTL in seconds.
    """

    ArchiveEnabled: bool = None
    EventTTL: int = None

    def __init__(self_, **kvargs):

        if "ArchiveEnabled" in kvargs and kvargs["ArchiveEnabled"] is not None:
            self_.ArchiveEnabled = kvargs["ArchiveEnabled"]
        if "EventTTL" in kvargs and kvargs["EventTTL"] is not None:
            self_.EventTTL = kvargs["EventTTL"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class HrefDeprecated:
    """

    Attributes
    ----------
    Href: str
    """

    Href: str = None

    def __init__(self_, **kvargs):

        if "Href" in kvargs and kvargs["Href"] is not None:
            self_.Href = kvargs["Href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ListLinks:
    """

    Attributes
    ----------
    Next: HrefDeprecated
    Prev: HrefDeprecated
    Self: HrefDeprecated
    next: Href
    prev: Href
    self: Href
    """

    Next: HrefDeprecated = None
    Prev: HrefDeprecated = None
    Self: HrefDeprecated = None
    next: Href = None
    prev: Href = None
    self: Href = None

    def __init__(self_, **kvargs):

        if "Next" in kvargs and kvargs["Next"] is not None:
            if type(kvargs["Next"]).__name__ == ListLinks.__annotations__["Next"]:
                self_.Next = kvargs["Next"]
            else:
                self_.Next = HrefDeprecated(**kvargs["Next"])
        if "Prev" in kvargs and kvargs["Prev"] is not None:
            if type(kvargs["Prev"]).__name__ == ListLinks.__annotations__["Prev"]:
                self_.Prev = kvargs["Prev"]
            else:
                self_.Prev = HrefDeprecated(**kvargs["Prev"])
        if "Self" in kvargs and kvargs["Self"] is not None:
            if type(kvargs["Self"]).__name__ == ListLinks.__annotations__["Self"]:
                self_.Self = kvargs["Self"]
            else:
                self_.Self = HrefDeprecated(**kvargs["Self"])
        if "next" in kvargs and kvargs["next"] is not None:
            if type(kvargs["next"]).__name__ == ListLinks.__annotations__["next"]:
                self_.next = kvargs["next"]
            else:
                self_.next = Href(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if type(kvargs["prev"]).__name__ == ListLinks.__annotations__["prev"]:
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Href(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == ListLinks.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = Href(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Href:
    """

    Attributes
    ----------
    href: str
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class Audits:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def get_archives(self, date: str) -> GetArchiveResult:
        """
        Retrieves audit events from long term storage.
        Finds and returns audit events from the archive, formatted as a JSON array, for the given date and tenant (in JWT).

        Parameters
        ----------
        date: str
          Date to be used as filter and criteria during extraction.
        """
        query_params = {}
        if date is not None:
            query_params["date"] = date
        response = self.auth.rest(
            path="/audits/archive",
            method="GET",
            params=query_params,
            data=None,
        )
        obj = GetArchiveResult(**response.json())
        obj.auth = self.auth
        return obj

    def get_settings(self) -> GetSettingsResult:
        """
        Returns the server configuration options.
        It includes options that represent the server configuration state and parameters that were used to run the server with certain functionality.

        Parameters
        ----------
        """
        response = self.auth.rest(
            path="/audits/settings",
            method="GET",
            params={},
            data=None,
        )
        obj = GetSettingsResult(**response.json())
        obj.auth = self.auth
        return obj

    def get_sources(self) -> GetObjectsResult:
        """
        Finds and returns the distinct list of unique event sources.

        Parameters
        ----------
        """
        response = self.auth.rest(
            path="/audits/sources",
            method="GET",
            params={},
            data=None,
        )
        obj = GetObjectsResult(**response.json())
        obj.auth = self.auth
        return obj

    def get_types(self) -> GetObjectsResult:
        """
        Finds and returns the distinct list of unique event types.

        Parameters
        ----------
        """
        response = self.auth.rest(
            path="/audits/types",
            method="GET",
            params={},
            data=None,
        )
        obj = GetObjectsResult(**response.json())
        obj.auth = self.auth
        return obj

    def gets(self, id: str) -> GetByIDResult:
        """
        Finds and returns the persisted audit events for the given tenant.

        Parameters
        ----------
        id: str
          The audit item's unique identifier.
        """
        response = self.auth.rest(
            path="/audits/{id}".replace("{id}", id),
            method="GET",
            params={},
            data=None,
        )
        obj = GetByIDResult(**response.json())
        obj.auth = self.auth
        return obj

    def get_audits(
        self,
        eventTime: str = None,
        eventType: str = None,
        id: str = None,
        limit: int = 10,
        next: str = None,
        prev: str = None,
        sort: str = "-eventTime",
        source: str = None,
        userId: str = None,
        max_items: int = 10,
    ) -> ListableResource[GetByIDResult]:
        """
        Finds and returns the persisted audit events for the given tenant.

        Parameters
        ----------
        eventTime: str = None
          The start/end time interval formatted in ISO 8601 to search by eventTime. For example, "?eventTime=2021-07-14T18:41:15.00Z/2021-07-14T18:41:15.99Z".
        eventType: str = None
          The case-sensitive string used to search by eventType.
        id: str = None
          The comma separated list of audit unique identifiers.
        limit: int = 10
          The maximum number of resources to return for a request. The limit must be an integer between 1 and 100 (inclusive).
        next: str = None
          The cursor to the next page of resources. Provide either the next or prev cursor, but not both.
        prev: str = None
          The cursor to the previous page of resources. Provide either the next or prev cursor, but not both.
        sort: str = "-eventTime"
          The property of a resource to sort on (default sort is -eventTime). The supported properties are source, eventType, and eventTime. A property must be prefixed by + or - to indicate ascending or descending sort order respectively.
        source: str = None
          The case-sensitive string used to search by source.
        userId: str = None
          The case-sensitive string used to search by userId.
        """
        query_params = {}
        if eventTime is not None:
            query_params["eventTime"] = eventTime
        if eventType is not None:
            query_params["eventType"] = eventType
        if id is not None:
            query_params["id"] = id
        if limit is not None:
            query_params["limit"] = limit
        if next is not None:
            query_params["next"] = next
        if prev is not None:
            query_params["prev"] = prev
        if sort is not None:
            query_params["sort"] = sort
        if source is not None:
            query_params["source"] = source
        if userId is not None:
            query_params["userId"] = userId
        response = self.auth.rest(
            path="/audits",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=GetByIDResult,
            auth=self.auth,
            path="/audits",
            max_items=max_items,
            query_params=query_params,
        )



# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import dataclass

from ..auth import Auth, Config


@dataclass
class GetQuotaByIdResult:
    """

    Attributes
    ----------
    data: list[Quota]
      Quota item.
    """

    data: list[Quota] = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == GetQuotaByIdResult.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [Quota(**e) for e in kvargs["data"]]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GetQuotasResult:
    """

    Attributes
    ----------
    data: list[Quota]
      Array of quota items.
    """

    data: list[Quota] = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]" == GetQuotasResult.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [Quota(**e) for e in kvargs["data"]]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Quota:
    """

    Attributes
    ----------
    attributes: QuotaAttributes
      The attributes of the quota.
    id: str
      The unique identifier of the quota item. For example, "app_mem_size", "app_upload_disk_size", or "shared_spaces".
    type: str
      The resource type of the quota item. Always equal to "quotas".
    """

    attributes: QuotaAttributes = None
    id: str = None
    type: str = None

    def __init__(self_, **kvargs):

        if "attributes" in kvargs and kvargs["attributes"] is not None:
            if (
                type(kvargs["attributes"]).__name__
                == Quota.__annotations__["attributes"]
            ):
                self_.attributes = kvargs["attributes"]
            else:
                self_.attributes = QuotaAttributes(**kvargs["attributes"])
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class QuotaAttributes:
    """
    The attributes of the quota.

    Attributes
    ----------
    quota: float
      The quota limit. If there is no quota limit, -1 is returned.
    unit: str
      The unit of the quota limit. For memory quotas, the unit is always "bytes". For other discrete units, the item counted is used as unit, for example "spaces".
    usage: float
      The current quota usage, if applicable. This attribute is only present if it is requested using the reportUsage query parameter.
    warningThresholds: list[float]
      The warning thresholds at which "close to quota" warnings can be issued when exceeded. If omitted, no warning threshold shall be used. Currently, the array will contain only one threshold value. In the future, this may be extended. The threshold is a number between 0 and 1, relating to the quota limit. For example, a value of 0.9 means that a warning should be issued when exceeding 90% of the quota limit.
    """

    quota: float = None
    unit: str = None
    usage: float = None
    warningThresholds: list[float] = None

    def __init__(self_, **kvargs):

        if "quota" in kvargs and kvargs["quota"] is not None:
            self_.quota = kvargs["quota"]
        if "unit" in kvargs and kvargs["unit"] is not None:
            self_.unit = kvargs["unit"]
        if "usage" in kvargs and kvargs["usage"] is not None:
            self_.usage = kvargs["usage"]
        if "warningThresholds" in kvargs and kvargs["warningThresholds"] is not None:
            self_.warningThresholds = kvargs["warningThresholds"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class Quotas:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def gets(self, id: str, reportUsage: bool = None) -> GetQuotaByIdResult:
        """
        Returns a specific quota item for the tenant (provided in JWT).

        Parameters
        ----------
        id: str
          The unique identifier of the quota item. For example, "app_mem_size", "app_upload_disk_size", or "shared_spaces".
        reportUsage: bool = None
          The Boolean flag indicating whether quota usage shall be part of the response. The default value is false (usage not included).
        """
        query_params = {}
        if reportUsage is not None:
            query_params["reportUsage"] = reportUsage
        response = self.auth.rest(
            path="/quotas/{id}".replace("{id}", id),
            method="GET",
            params=query_params,
            data=None,
        )
        obj = GetQuotaByIdResult(**response.json())
        obj.auth = self.auth
        return obj

    def get_quotas(self, reportUsage: bool = None) -> GetQuotasResult:
        """
        Returns all quota items for the tenant (provided in JWT).

        Parameters
        ----------
        reportUsage: bool = None
          The Boolean flag indicating whether quota usage shall be part of the response. The default value is false (only limits returned).
        """
        query_params = {}
        if reportUsage is not None:
            query_params["reportUsage"] = reportUsage
        response = self.auth.rest(
            path="/quotas",
            method="GET",
            params=query_params,
            data=None,
        )
        obj = GetQuotasResult(**response.json())
        obj.auth = self.auth
        return obj



from enum import Enum


class CustomException(Exception):
    """
    CustomException represents a custom exception
    """


class CustomExceptionMsg(Enum):
    """
    CustomExceptionMsg - messages for custom exceptions
    """

    EMPTY_HOST = "Empty host in config"
    MISSING_CONFIG_PROPERTY = "Missing config property"
    UNSUPPORTED_AUTH_TYPE = "Unsupported authType"
    UNSUPPORTED_PROPERTY = "Unsupported property"
    NOT_IMPLEMENTED = "Not implemented"


class AuthorizeException(CustomException):
    error: str
    error_code: str
    error_description: str
    error_detail: str
    error_uri: str

    def __init__(self, *args: object) -> None:
        self.error = None
        self.error_code = None
        self.error_description = None
        self.error_detail = None
        self.error_uri = None
        super().__init__(*args)
        for a in args:
            for key in self.__dict__.keys():
                if key in a:
                    setattr(self, key, a[key])


class EngineClosedException(Exception):
    close_code: str
    close_msg: str

    def __init__(self, *args: object) -> None:
        self.close_code = None
        self.close_msg = None
        super().__init__(*args)
        for a in args:
            for key in self.__dict__.keys():
                if key in a:
                    setattr(self, key, a[key])



from typing import Generic, List, TypeVar, Union

T = TypeVar("T")
ResI = TypeVar("ResI")
ReqI = TypeVar("ReqI")


class InterceptorHandler(Generic[T]):
    handlers: List[T]
    "list containing the interceptors"

    def __init__(self):
        self.handlers = []

    def use(self, interceptor: Union[T, List[T]]) -> None:
        """
        method helper for registering an interceptor

        Parameters
        ----------
        interceptor
            function interceptor for requests/responses
        """
        if isinstance(interceptor, list):
            self.handlers = self.handlers + interceptor
        else:
            self.handlers.append(interceptor)


class Interceptors(Generic[ReqI, ResI]):
    response: InterceptorHandler[ResI]
    request: InterceptorHandler[ReqI]



from .apis import (
    Api_Keys,
    Apps,
    Audits,
    Automations,
    Collections,
    Csp_Origins,
    Data_Files,
    Extensions,
    Groups,
    Identity_Providers,
    Items,
    Licenses,
    Oauth_Tokens,
    Questions,
    Quotas,
    Reload_Tasks,
    Reloads,
    Roles,
    Spaces,
    Themes,
    Users,
    Web_Integrations,
    Webhooks,
)
from .auth import Auth
from .config import Config
from .rest import RestClientInstance
from .rpc import RpcClientInstance


class Qlik:
    """
    Qlik Class is the entry-point for the Qlik python Platform SDK

    Parameters
    ----------
    config: Config
        the required configuration object

    Examples
    --------
    >>> from qlik_sdk import Qlik
    ...
    ... qlik = Qlik(Config(host=base_url, auth_type=AuthType.APIKey, api_key=api_key))
    ... user_me = qlik.users.get_me()
    """

    config: Config
    auth: Auth
    rpc: RpcClientInstance
    """
    rpc returns an RpcClient that can be used to
    connect to the engine for a specific app

    Parameters
    ----------
    app_id: str

    Attributes
    ----------
    interceptors: Interceptors

    Examples
    ----------
    >>> rpc_session = auth.rpc(app_id=session_app_id)
    ... with rpc_session.opn() as rpc_client:
    ...     app = rpc_client.send("OpenDoc", -1, session_app_id)
    # And with interceptors.
    >>> auth.rpc.interceptors["request"].use(log_req)
    ... rpc_session = auth.rpc(app_id=session_app_id)
    ...
    ... with rpc_session.open() as rpc_client:
    ...     app = rpc_client.send("OpenDoc", -1, session_app_id)
    """

    rest: RestClientInstance
    """
    rest method can be used to make raw calls against Qlik Cloud

    Parameters
    ----------
    app_id: str
    method: str
        HTTP verb default GET
    path: str
        represents the api endpoint ex: `/users/me`
    data: dict, optional
        Dictionary, list of tuples, bytes, or file-like object to send in the body of the Request.
    params: dict, optional
        Dictionary, list of tuples or bytes to send in the query string for the Request.
    headers: dict, optional
        Dictionary of HTTP Headers to send with the Request
    files: dict, optional
        Dictionary of {filename: fileobject} files to multipart upload.

    Attributes
    ----------
    interceptors: Interceptors

    Examples
    ----------
    >>> auth = Auth(Config(host=self.base_url, auth_type=AuthType.APIKey, api_key=self.api_key))
    ... user_me = auth.rest(path="/users/me")
    # And with interceptors.
    >>> auth = Auth(Config(host=self.base_url, auth_type=AuthType.APIKey, api_key=self.api_key))
    ... def log_req(req: requests.Request) -> requests.Request:
    ...     print(req)
    ...     return req
    ...
    ... auth.rpc.interceptors["request"].use(log_req)
    ... app_list = auth.rest(path="/items", params={"resourceType":"app", "limit": 100})
    """

    apps: Apps.Apps
    items: Items.Items
    spaces: Spaces.Spaces
    users: Users.Users
    extensions: Extensions.Extensions
    reloads: Reloads.Reloads
    themes: Themes.Themes
    datafiles: Data_Files.DataFiles
    roles: Roles.Roles
    groups: Groups.Groups
    licenses: Licenses.Licenses
    identityproviders: Identity_Providers.IdentityProviders
    csporigins: Csp_Origins.CspOrigins
    webintegrations: Web_Integrations.WebIntegrations
    apikeys: Api_Keys.ApiKeys
    audits: Audits.Audits
    quotas: Quotas.Quotas
    webhooks: Webhooks.Webhooks
    automations: Automations.Automations
    collections: Collections.Collections
    reload_tasks: Reload_Tasks.ReloadTasks
    oauth_tokens: Oauth_Tokens.OauthTokens
    questions: Questions.Questions

    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config=config)
        self.rest = self.auth.rest
        self.rpc = self.auth.rpc
        self.apps = Apps.Apps(config=config)
        self.extensions = Extensions.Extensions(config=config)
        self.items = Items.Items(config=config)
        self.spaces = Spaces.Spaces(config=config)
        self.users = Users.Users(config=config)
        self.reloads = Reloads.Reloads(config=config)
        self.themes = Themes.Themes(config=config)
        self.datafiles = Data_Files.DataFiles(config=config)
        self.roles = Roles.Roles(config=config)
        self.groups = Groups.Groups(config=config)
        self.licenses = Licenses.Licenses(config=config)
        self.identityproviders = Identity_Providers.IdentityProviders(config=config)
        self.csporigins = Csp_Origins.CspOrigins(config=config)
        self.webintegrations = Web_Integrations.WebIntegrations(config=config)
        self.apikeys = Api_Keys.ApiKeys(config=config)
        self.audits = Audits.Audits(config=config)
        self.quotas = Quotas.Quotas(config=config)
        self.webhooks = Webhooks.Webhooks(config=config)
        self.automations = Automations.Automations(config=config)
        self.collections = Collections.Collections(config=config)
        self.reload_tasks = Reload_Tasks.ReloadTasks(config=config)
        self.oauth_tokens = Oauth_Tokens.OauthTokens(config=config)
        self.questions = Questions.Questions(config=config)



# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import List, Literal

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class ApiKey:
    """

    Attributes
    ----------
    created: str
      When the API key was created.
    createdByUser: str
      The ID of the user who created the key.
    description: str
      A description for the API key.
    expiry: str
      When the API key will expire and no longer be a valid authentication token.
    id: str
      The unique ID for the resource.
    lastUpdated: str
      When the API key was last updated.
    status: Literal["active", "expired", "revoked"]
    sub: str
      The ID of the subject for the API key.
    subType: Literal["user"]
      Type of the subject.
    tenantId: str
      The tenant ID.
    """

    created: str = None
    createdByUser: str = None
    description: str = None
    expiry: str = None
    id: str = None
    lastUpdated: str = None
    status: Literal["active", "expired", "revoked"] = None
    sub: str = None
    subType: Literal["user"] = None
    tenantId: str = None

    def __init__(self_, **kvargs):

        if "created" in kvargs and kvargs["created"] is not None:
            self_.created = kvargs["created"]
        if "createdByUser" in kvargs and kvargs["createdByUser"] is not None:
            self_.createdByUser = kvargs["createdByUser"]
        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "expiry" in kvargs and kvargs["expiry"] is not None:
            self_.expiry = kvargs["expiry"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "lastUpdated" in kvargs and kvargs["lastUpdated"] is not None:
            self_.lastUpdated = kvargs["lastUpdated"]
        if "status" in kvargs and kvargs["status"] is not None:
            self_.status = kvargs["status"]
        if "sub" in kvargs and kvargs["sub"] is not None:
            self_.sub = kvargs["sub"]
        if "subType" in kvargs and kvargs["subType"] is not None:
            self_.subType = kvargs["subType"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)

    def delete(self) -> None:
        """
        Deletes or revokes an API key.
        When the owner of the API key sends the request, the key will be removed. When a TenantAdmin sends the request, the key will be revoked.

        Parameters
        ----------
        """
        self.auth.rest(
            path="/api-keys/{id}".replace("{id}", self.id),
            method="DELETE",
            params={},
            data=None,
        )

    def patch(self, data: ApiKeysPatchSchema) -> None:
        """
        Updates an API key for a given ID.

        Parameters
        ----------
        data: ApiKeysPatchSchema

        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        self.auth.rest(
            path="/api-keys/{id}".replace("{id}", self.id),
            method="PATCH",
            params={},
            data=data,
        )


@dataclass
class ApiKeyBody:
    """

    Attributes
    ----------
    description: str
      Text that describes the API key.
    expiry: str
      The expiry of the API key, in ISO8601 duration format.
    sub: str
      The ID of the subject for the API key.
    subType: str
      Type of the subject.
    """

    description: str = None
    expiry: str = None
    sub: str = None
    subType: str = None

    def __init__(self_, **kvargs):

        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "expiry" in kvargs and kvargs["expiry"] is not None:
            self_.expiry = kvargs["expiry"]
        if "sub" in kvargs and kvargs["sub"] is not None:
            self_.sub = kvargs["sub"]
        if "subType" in kvargs and kvargs["subType"] is not None:
            self_.subType = kvargs["subType"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ApiKeyConfigPatch:
    """
    A JSON Patch document as defined in https://datatracker.ietf.org/doc/html/rfc6902.

    Attributes
    ----------
    op: Literal["replace"]
      The operation to be performed.
    path: Literal["/api_keys_enabled", "/max_api_key_expiry", "/max_keys_per_user"]
      The path for the given resource field to patch.
    value: object
      The value to be used for this operation.
    """

    op: Literal["replace"] = None
    path: Literal[
        "/api_keys_enabled", "/max_api_key_expiry", "/max_keys_per_user"
    ] = None
    value: object = None

    def __init__(self_, **kvargs):

        if "op" in kvargs and kvargs["op"] is not None:
            self_.op = kvargs["op"]
        if "path" in kvargs and kvargs["path"] is not None:
            self_.path = kvargs["path"]
        if "value" in kvargs and kvargs["value"] is not None:
            self_.value = kvargs["value"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ApiKeyPatch:
    """
    A JSON Patch document as defined in https://datatracker.ietf.org/doc/html/rfc6902.

    Attributes
    ----------
    op: Literal["replace"]
      The operation to be performed.
    path: Literal["/description"]
      The path for the given resource field to patch.
    value: str
      The value to be used for this operation.
    """

    op: Literal["replace"] = None
    path: Literal["/description"] = None
    value: str = None

    def __init__(self_, **kvargs):

        if "op" in kvargs and kvargs["op"] is not None:
            self_.op = kvargs["op"]
        if "path" in kvargs and kvargs["path"] is not None:
            self_.path = kvargs["path"]
        if "value" in kvargs and kvargs["value"] is not None:
            self_.value = kvargs["value"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ApiKeyWithToken:
    """

    Attributes
    ----------
    created: str
      When the API key was created.
    createdByUser: str
      The id of the user who created the key.
    description: str
      A description for the API key.
    expiry: str
      When the API key will expire and no longer be a valid authentication token.
    id: str
      The unique ID for the resource.
    lastUpdated: str
      When the API key was last updated.
    status: Literal["active", "expired", "revoked"]
    sub: str
      The ID of the subject for the API key.
    subType: Literal["user"]
      Type of the subject.
    tenantId: str
      The tenant ID.
    token: str
      The generated signed JWT.
    """

    created: str = None
    createdByUser: str = None
    description: str = None
    expiry: str = None
    id: str = None
    lastUpdated: str = None
    status: Literal["active", "expired", "revoked"] = None
    sub: str = None
    subType: Literal["user"] = None
    tenantId: str = None
    token: str = None

    def __init__(self_, **kvargs):

        if "created" in kvargs and kvargs["created"] is not None:
            self_.created = kvargs["created"]
        if "createdByUser" in kvargs and kvargs["createdByUser"] is not None:
            self_.createdByUser = kvargs["createdByUser"]
        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "expiry" in kvargs and kvargs["expiry"] is not None:
            self_.expiry = kvargs["expiry"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "lastUpdated" in kvargs and kvargs["lastUpdated"] is not None:
            self_.lastUpdated = kvargs["lastUpdated"]
        if "status" in kvargs and kvargs["status"] is not None:
            self_.status = kvargs["status"]
        if "sub" in kvargs and kvargs["sub"] is not None:
            self_.sub = kvargs["sub"]
        if "subType" in kvargs and kvargs["subType"] is not None:
            self_.subType = kvargs["subType"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "token" in kvargs and kvargs["token"] is not None:
            self_.token = kvargs["token"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ApiKeysConfig:
    """

    Attributes
    ----------
    api_keys_enabled: bool
      Enables or disables API key functionality for the specified tenant.
    max_api_key_expiry: str
      The maximum lifetime, in ISO8601 duration format, for which an API key can be issued for the specified tenant.
    max_keys_per_user: float
      The maximum number of active API keys that any user can create for the specified tenant.
    scim_externalClient_expiry: str
      The expiry of the scim externalClient token in ISO8601 duration format. Used during the creation of an externalClient API key.
    """

    api_keys_enabled: bool = None
    max_api_key_expiry: str = "PT24H"
    max_keys_per_user: float = 5
    scim_externalClient_expiry: str = "P365D"

    def __init__(self_, **kvargs):

        if "api_keys_enabled" in kvargs and kvargs["api_keys_enabled"] is not None:
            self_.api_keys_enabled = kvargs["api_keys_enabled"]
        if "max_api_key_expiry" in kvargs and kvargs["max_api_key_expiry"] is not None:
            self_.max_api_key_expiry = kvargs["max_api_key_expiry"]
        if "max_keys_per_user" in kvargs and kvargs["max_keys_per_user"] is not None:
            self_.max_keys_per_user = kvargs["max_keys_per_user"]
        if (
            "scim_externalClient_expiry" in kvargs
            and kvargs["scim_externalClient_expiry"] is not None
        ):
            self_.scim_externalClient_expiry = kvargs["scim_externalClient_expiry"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ApiKeysConfigPatchSchema(List["ApiKeyConfigPatch"]):
    """

    Attributes
    ----------
    """

    def __init__(self_, elements):
        for e in elements:
            self_.append(ApiKeyConfigPatch(**e))


@dataclass
class ApiKeysPatchSchema(List["ApiKeyPatch"]):
    """

    Attributes
    ----------
    """

    def __init__(self_, elements):
        for e in elements:
            self_.append(ApiKeyPatch(**e))


@dataclass
class Link:
    """

    Attributes
    ----------
    href: str
      The URL for the link.
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ApiKeyPage:
    """

    Attributes
    ----------
    data: list[ApiKey]
    links: ApiKeyPageLinks
    """

    data: list[ApiKey] = None
    links: ApiKeyPageLinks = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]" == ApiKeyPage.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [ApiKey(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == ApiKeyPage.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = ApiKeyPageLinks(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ApiKeyPageLinks:
    """

    Attributes
    ----------
    next: Link
    prev: Link
    self: Link
    """

    next: Link = None
    prev: Link = None
    self: Link = None

    def __init__(self_, **kvargs):

        if "next" in kvargs and kvargs["next"] is not None:
            if type(kvargs["next"]).__name__ == ApiKeyPageLinks.__annotations__["next"]:
                self_.next = kvargs["next"]
            else:
                self_.next = Link(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if type(kvargs["prev"]).__name__ == ApiKeyPageLinks.__annotations__["prev"]:
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Link(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == ApiKeyPageLinks.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class ApiKeys:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def get_config(self, tenantId: str) -> ApiKeysConfig:
        """
        Gets the API keys configuration for a given tenant ID.

        Parameters
        ----------
        tenantId: str
          The tenant ID of the API keys configuration to be retrieved.
        """
        response = self.auth.rest(
            path="/api-keys/configs/{tenantId}".replace("{tenantId}", tenantId),
            method="GET",
            params={},
            data=None,
        )
        obj = ApiKeysConfig(**response.json())
        obj.auth = self.auth
        return obj

    def patch_config(self, tenantId: str, data: ApiKeysConfigPatchSchema) -> None:
        """
        Updates the API keys configuration for a given tenant ID.

        Parameters
        ----------
        tenantId: str
          The tenant ID of the API keys configuration to be retrieved.
        data: ApiKeysConfigPatchSchema

        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        self.auth.rest(
            path="/api-keys/configs/{tenantId}".replace("{tenantId}", tenantId),
            method="PATCH",
            params={},
            data=data,
        )

    def get(self, id: str) -> ApiKey:
        """
        Gets the API key for a given ID.

        Parameters
        ----------
        id: str
          The ID of the API key resource to be retrieved.
        """
        response = self.auth.rest(
            path="/api-keys/{id}".replace("{id}", id),
            method="GET",
            params={},
            data=None,
        )
        obj = ApiKey(**response.json())
        obj.auth = self.auth
        return obj

    def get_api_keys(
        self,
        createdByUser: str = None,
        endingBefore: str = None,
        limit: float = 20,
        sort: Literal[
            "createdByUser",
            "+createdByUser",
            "-createdByUser",
            "sub",
            "+sub",
            "-sub",
            "status",
            "+status",
            "-status",
            "description",
            "+description",
            "-description",
            "created",
            "+created",
            "-created",
        ] = "-created",
        startingAfter: str = None,
        status: Literal["active", "expired", "revoked"] = None,
        sub: str = None,
        max_items: int = 20,
    ) -> ListableResource[ApiKey]:
        """
        Lists API keys for a given tenant ID.

        Parameters
        ----------
        createdByUser: str = None
          The user ID that created the API key.
        endingBefore: str = None
          Get resources with IDs that are lower than the target resource ID. Cannot be used in conjunction with startingAfter.
        limit: float = 20
          Maximum number of API keys to retrieve.
        sort: Literal["createdByUser", "+createdByUser", "-createdByUser", "sub", "+sub", "-sub", "status", "+status", "-status", "description", "+description", "-description", "created", "+created", "-created"] = "-created"
          The field to sort by, with +/- prefix indicating sort order
        startingAfter: str = None
          Get resources with IDs that are higher than the target resource ID. Cannot be used in conjunction with endingBefore.
        status: Literal["active", "expired", "revoked"] = None
          The status of the API key.
        sub: str = None
          The ID of the subject.
        """
        query_params = {}
        if createdByUser is not None:
            query_params["createdByUser"] = createdByUser
        if endingBefore is not None:
            query_params["endingBefore"] = endingBefore
        if limit is not None:
            query_params["limit"] = limit
        if sort is not None:
            query_params["sort"] = sort
        if startingAfter is not None:
            query_params["startingAfter"] = startingAfter
        if status is not None:
            query_params["status"] = status
        if sub is not None:
            query_params["sub"] = sub
        response = self.auth.rest(
            path="/api-keys",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=ApiKey,
            auth=self.auth,
            path="/api-keys",
            max_items=max_items,
            query_params=query_params,
        )

    def create(self, data: ApiKeyBody) -> ApiKeyWithToken:
        """
        Creates an API key resource.

        Parameters
        ----------
        data: ApiKeyBody

        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/api-keys",
            method="POST",
            params={},
            data=data,
        )
        obj = ApiKeyWithToken(**response.json())
        obj.auth = self.auth
        return obj



import typing
from urllib.parse import parse_qs, urlparse

from .auth_type import AuthType
from .config import Config
from .errors import AuthorizeException, CustomException
from .oauth_utils import _generate_authorization_url
from .rest import RestClient, RestClientInstance
from .rpc import RpcClient, RpcClientInstance


class Auth:
    """
    Auth can be used to make rest and rpc calls

    Parameters
    ----------
    config: Config
        the required configuration object

    Examples
    --------
    >>> from qlik_sdk import Auth, AuthType, Config
    ...
    ... clients = Auth(Config(host=base_url, auth_type=AuthType.APIKey, api_key=api_key))
    ... get_users_res = clients.rest(path="/users/me")
    """

    config: Config
    rest: RestClientInstance
    """
    rest method can be used to make raw calls against Qlik Cloud

    Parameters
    ----------
    method: str, default GET
        string HTTP verb
    path: str
        representing the api endpoint ex: `/users/me`
    data: dict, optional
        Dictionary, list of tuples, bytes, or file-like object to send in the body of the Request.
    params: dict, optional
        Dictionary, list of tuples or bytes to send in the query string for the Request.
    files: dict, optional
        Dictionary of {filename: fileobject} files to multipart upload.
    headers: dict, optional
        Dictionary of HTTP Headers to send with the Request
    stream: bool, optional, default True
        if False, the response content will be immediately downloaded.
    timeout: int optional, default 10
        How many seconds to wait for the server to send data before giving up

    Attributes
    ----------
    interceptors: Interceptors

    Examples
    ----------
    >>> auth = Auth(Config(host=self.base_url, auth_type=AuthType.APIKey, api_key=self.api_key))
    ... user_me = auth.rest(path="/users/me")
    ...
    # And with interceptors.
    >>> auth = Auth(Config(host=self.base_url, auth_type=AuthType.APIKey, api_key=self.api_key))
    ... def log_req(req: requests.Request) -> requests.Request:
    ...     print(req)
    ...     return req
    ...
    ... auth.rpc.interceptors["request"].use(log_req)
    ... app_list = auth.rest(path="/items", params={"resourceType":"app", "limit": 100})
    """

    rpc: RpcClientInstance
    """
    rpc returns an RpcClient that can be used to
    connect to the engine for a specific app

    Parameters
    ----------
    app_id: str

    Attributes
    ----------
    interceptors: Interceptors

    Examples
    ----------
    >>> rpc_session = auth.rpc(app_id=session_app_id)
    ... with rpc_session.opn() as rpc_client:
    ...     app = rpc_client.send("OpenDoc", -1, session_app_id)
    ...
    # And with interceptors.
    >>> auth.rpc.interceptors["request"].use(log_req)
    ... rpc_session = auth.rpc(app_id=session_app_id)
    ...
    ... with rpc_session.open() as rpc_client:
    ...     app = rpc_client.send("OpenDoc", -1, session_app_id)
    """
    paginate_rest: typing.Callable[[str, str, typing.Dict, any], typing.Iterator[any]]

    code_verifier: str = None
    """oauth code verifier"""
    state: str = None
    """oauth state"""

    def __init__(self, config: Config):
        config.validate()
        self.config = config
        self.rest = RestClientInstance(RestClient(config))
        self.rpc_client = RpcClient(self.config)
        self.rpc = RpcClientInstance(self.rpc_client)
        self.paginate_rest = self.rest.paginate_rest

    def generate_authorization_url(self, state: str = None) -> str:
        """
        method helper for generating the authorization url in OAuth flows based on the Config

        Parameters
        ----------
        state: str, optional
            the state parameter is a random generated string used for validating against
            potential CSRF attacks

        Examples
        ----------
        >>> auth = Auth(Config(
        ...     host=host,
        ...     auth_type = AuthType.OAuth2,
        ...     client_id = <...>,
        ...     client_secret = <...>,
        ...     redirect_url = <...>))
        ...
        ... auth_url = auth.generate_authorization_url()
        ... webbrowser.get('chrome').open(url)
        """
        data = _generate_authorization_url(self.config, state, True)
        self.code_verifier = data["code_verifier"]
        self.state = data["state"]
        return data["url"]

    def authorize(self, url: str = None) -> None:
        """
        exchange credentials with token against the authorization server

        Parameters
        ----------
        url: str, optional
            The callback url with query parameters
        """
        code = None
        state = None
        error = None
        query = None
        if url is not None:
            try:
                query = parse_qs(urlparse(url).query)
            except:
                pass
        if query is not None:
            if "code" in query and len(query["code"]) > 0:
                code = query["code"][0]
            if "state" in query and len(query["state"]) > 0:
                state = query["state"][0]
            if "error" in query and len(query["error"]) > 0:
                error = query["error"][0]

        if error is not None:
            error_dict = {}
            for key in [
                "error",
                "error_code",
                "error_description",
                "error_detail",
                "error_uri",
            ]:
                val = query.get(key, None)
                if val is not None:
                    error_dict[key] = val[0]
                else:
                    error_dict[key] = None
            raise AuthorizeException(error_dict)
        if self.state is not None and state != self.state:
            raise CustomException(
                "The received state parameter does not match the original state value"
            )
        params = {
            "grant_type": "client_credentials",
            "scope": " ".join(self.config.scope),
        }
        if code is not None:
            params["code"] = code
            params["grant_type"] = "authorization_code"
            params["code_verifier"] = self.code_verifier
            params["client_id"] = self.config.client_id
        if self.config.redirect_url is not None and self.config.redirect_url != "":
            params["redirect_uri"] = self.config.redirect_url
        resp = self.rest(path="/oauth/token", method="post", data=params)
        resp = resp.json()
        self.config.api_key = resp["access_token"]
        if "refresh_token" in resp:
            self.config.refresh_token = resp["refresh_token"]
        # return response from /oauth/token
        return resp

    def refresh_token(self) -> None:
        """
        Requesting new access_token using refresh token on expiry of current access_token
        """
        if (
            self.config.auth_type != AuthType.OAuth2
            or "offline_access" not in self.config.scope
            or not self.config.refresh_token
        ):
            raise CustomException(
                "Method only available for AuthType OAuth2, that have offline_access scope"
            )
        params = {
            "grant_type": "refresh_token",
            "refresh_token": self.config.refresh_token,
        }
        resp = self.rest(path="/oauth/token", method="post", data=params, headers={})
        resp = resp.json()
        self.config.api_key = resp["access_token"]
        self.config.refresh_token = resp["refresh_token"]

    def deauthorize(self) -> None:
        """
        deauthorize
        revokes the Auth instance oauth token
        """
        body = {"token": self.config.api_key, "token_type_hint": "access_token"}
        self.rest(path="/oauth/revoke", method="post", data=body)
        self.config.api_key = ""
        self.config.refresh_token = ""



# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

import warnings
from dataclasses import asdict, dataclass
from typing import Literal

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class Space:
    """
    A space is a security context simplifying the management of access control by allowing users to control it on the containers instead of on the resources themselves.

    Attributes
    ----------
    createdAt: str
      The date and time when the space was created.
    createdBy: str
      The ID of the user who created the space.
    description: str
      The description of the space. Personal spaces do not have a description.
    id: str
      A unique identifier for the space, for example, 62716f4b39b865ece543cd45.
    links: SpaceLinks
    meta: SpaceMeta
      Information about the space settings.
    name: str
      The name of the space. Personal spaces do not have a name.
    ownerId: str
      The ID for the space owner.
    tenantId: str
      The ID for the tenant, for example, xqGQ0k66vSR8f9G7J-vYtHZQkiYrCpct.
    type: Literal["shared", "managed", "data"]
      The type of space such as shared, managed, and so on.
    updatedAt: str
      The date and time when the space was updated.
    """

    createdAt: str = None
    createdBy: str = None
    description: str = None
    id: str = None
    links: SpaceLinks = None
    meta: SpaceMeta = None
    name: str = None
    ownerId: str = None
    tenantId: str = None
    type: Literal["shared", "managed", "data"] = None
    updatedAt: str = None

    def __init__(self_, **kvargs):

        if "createdAt" in kvargs and kvargs["createdAt"] is not None:
            self_.createdAt = kvargs["createdAt"]
        if "createdBy" in kvargs and kvargs["createdBy"] is not None:
            self_.createdBy = kvargs["createdBy"]
        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == Space.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = SpaceLinks(**kvargs["links"])
        if "meta" in kvargs and kvargs["meta"] is not None:
            if type(kvargs["meta"]).__name__ == Space.__annotations__["meta"]:
                self_.meta = kvargs["meta"]
            else:
                self_.meta = SpaceMeta(**kvargs["meta"])
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "ownerId" in kvargs and kvargs["ownerId"] is not None:
            self_.ownerId = kvargs["ownerId"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        if "updatedAt" in kvargs and kvargs["updatedAt"] is not None:
            self_.updatedAt = kvargs["updatedAt"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)

    def delete_assignment(self, assignmentId: str) -> None:
        """
        Deletes an assignment.

        Parameters
        ----------
        assignmentId: str
          The ID of the assignment to delete.
        """
        self.auth.rest(
            path="/spaces/{spaceId}/assignments/{assignmentId}".replace(
                "{assignmentId}", assignmentId
            ).replace("{spaceId}", self.id),
            method="DELETE",
            params={},
            data=None,
        )

    def get_assignment(self, assignmentId: str) -> Assignment:
        """
        Retrieves a single assignment by ID.

        Parameters
        ----------
        assignmentId: str
          The ID of the assignment to retrieve.
        """
        response = self.auth.rest(
            path="/spaces/{spaceId}/assignments/{assignmentId}".replace(
                "{assignmentId}", assignmentId
            ).replace("{spaceId}", self.id),
            method="GET",
            params={},
            data=None,
        )
        obj = Assignment(**response.json())
        obj.auth = self.auth
        return obj

    def set_assignment(self, assignmentId: str, data: AssignmentUpdate) -> Assignment:
        """
        Updates a single assignment by ID. The complete list of roles must be provided.

        Parameters
        ----------
        assignmentId: str
          The ID of the assignment to update.
        data: AssignmentUpdate
          Attributes that the user wants to update for the specified assignment.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/spaces/{spaceId}/assignments/{assignmentId}".replace(
                "{assignmentId}", assignmentId
            ).replace("{spaceId}", self.id),
            method="PUT",
            params={},
            data=data,
        )
        obj = Assignment(**response.json())
        obj.auth = self.auth
        return obj

    def get_assignments(
        self, limit: int = 10, next: str = None, prev: str = None, max_items: int = 10
    ) -> ListableResource[Assignment]:
        """
        Retrieves the assignments of the space matching the query.

        Parameters
        ----------
        limit: int = 10
          Maximum number of assignments to return.
        next: str = None
          The next page cursor. Next links make use of this.
        prev: str = None
          The previous page cursor. Previous links make use of this.
        """
        query_params = {}
        if limit is not None:
            query_params["limit"] = limit
        if next is not None:
            query_params["next"] = next
        if prev is not None:
            query_params["prev"] = prev
        response = self.auth.rest(
            path="/spaces/{spaceId}/assignments".replace("{spaceId}", self.id),
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=Assignment,
            auth=self.auth,
            path="/spaces/{spaceId}/assignments".replace("{spaceId}", self.id),
            max_items=max_items,
            query_params=query_params,
        )

    def create_assignment(self, data: AssignmentCreate) -> Assignment:
        """
        Creates an assignment.

        Parameters
        ----------
        data: AssignmentCreate
          Attributes that the user wants to set for the assignment for the space.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/spaces/{spaceId}/assignments".replace("{spaceId}", self.id),
            method="POST",
            params={},
            data=data,
        )
        obj = Assignment(**response.json())
        obj.auth = self.auth
        return obj

    def delete(self) -> None:
        """
        Deletes a space.

        Parameters
        ----------
        """
        self.auth.rest(
            path="/spaces/{spaceId}".replace("{spaceId}", self.id),
            method="DELETE",
            params={},
            data=None,
        )

    def patch(self, data: SpacePatch) -> Space:
        """
        Patches (updates) a space (partially).

        Parameters
        ----------
        data: SpacePatch
          Attribute that the user wants to patch (update) for the specified space.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/spaces/{spaceId}".replace("{spaceId}", self.id),
            method="PATCH",
            params={},
            data=data,
        )
        self.__init__(**response.json())
        return self

    def set(self, data: SpaceUpdate) -> Space:
        """
        Updates a space.

        Parameters
        ----------
        data: SpaceUpdate
          Attributes that the user wants to update for the specified space.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/spaces/{spaceId}".replace("{spaceId}", self.id),
            method="PUT",
            params={},
            data=data,
        )
        self.__init__(**response.json())
        return self


@dataclass
class Assignment:
    """

    Attributes
    ----------
    assigneeId: str
      The userId or groupId based on the type.
    createdAt: str
      The date and time when the space was created.
    createdBy: str
      The ID of the user who created the assignment.
    id: str
    links: AssignmentLinks
    roles: list[str]
      The roles assigned to a user or group. Must not be empty.
    spaceId: str
      The unique identifier for the space.
    tenantId: str
      The unique identifier for the tenant.
    type: Literal["user", "group"]
    updatedAt: str
      The date and time when the space was updated.
    updatedBy: str
      The ID of the user who updated the assignment.
    """

    assigneeId: str = None
    createdAt: str = None
    createdBy: str = None
    id: str = None
    links: AssignmentLinks = None
    roles: list[str] = None
    spaceId: str = None
    tenantId: str = None
    type: Literal["user", "group"] = None
    updatedAt: str = None
    updatedBy: str = None

    def __init__(self_, **kvargs):

        if "assigneeId" in kvargs and kvargs["assigneeId"] is not None:
            self_.assigneeId = kvargs["assigneeId"]
        if "createdAt" in kvargs and kvargs["createdAt"] is not None:
            self_.createdAt = kvargs["createdAt"]
        if "createdBy" in kvargs and kvargs["createdBy"] is not None:
            self_.createdBy = kvargs["createdBy"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == Assignment.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = AssignmentLinks(**kvargs["links"])
        if "roles" in kvargs and kvargs["roles"] is not None:
            self_.roles = kvargs["roles"]
        if "spaceId" in kvargs and kvargs["spaceId"] is not None:
            self_.spaceId = kvargs["spaceId"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        if "updatedAt" in kvargs and kvargs["updatedAt"] is not None:
            self_.updatedAt = kvargs["updatedAt"]
        if "updatedBy" in kvargs and kvargs["updatedBy"] is not None:
            self_.updatedBy = kvargs["updatedBy"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentCreate:
    """

    Attributes
    ----------
    assigneeId: str
      The userId or groupId based on the type.
    roles: list[str]
      The roles assigned to the assigneeId.
    type: Literal["user", "group"]
      The type of space such as shared, managed, and so on.
    """

    assigneeId: str = None
    roles: list[str] = None
    type: Literal["user", "group"] = None

    def __init__(self_, **kvargs):

        if "assigneeId" in kvargs and kvargs["assigneeId"] is not None:
            self_.assigneeId = kvargs["assigneeId"]
        if "roles" in kvargs and kvargs["roles"] is not None:
            self_.roles = kvargs["roles"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentLinks:
    """

    Attributes
    ----------
    self: Link
    space: Link
    """

    self: Link = None
    space: Link = None

    def __init__(self_, **kvargs):

        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == AssignmentLinks.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        if "space" in kvargs and kvargs["space"] is not None:
            if (
                type(kvargs["space"]).__name__
                == AssignmentLinks.__annotations__["space"]
            ):
                self_.space = kvargs["space"]
            else:
                self_.space = Link(**kvargs["space"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentUpdate:
    """

    Attributes
    ----------
    roles: list[str]
      The roles assigned to the assigneeId.
    """

    roles: list[str] = None

    def __init__(self_, **kvargs):

        if "roles" in kvargs and kvargs["roles"] is not None:
            self_.roles = kvargs["roles"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Assignments:
    """

    Attributes
    ----------
    data: list[Assignment]
    links: AssignmentsLinks
    meta: AssignmentsMeta
    """

    data: list[Assignment] = None
    links: AssignmentsLinks = None
    meta: AssignmentsMeta = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]" == Assignments.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [Assignment(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == Assignments.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = AssignmentsLinks(**kvargs["links"])
        if "meta" in kvargs and kvargs["meta"] is not None:
            if type(kvargs["meta"]).__name__ == Assignments.__annotations__["meta"]:
                self_.meta = kvargs["meta"]
            else:
                self_.meta = AssignmentsMeta(**kvargs["meta"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsLinks:
    """

    Attributes
    ----------
    next: Link
    prev: Link
    self: Link
    """

    next: Link = None
    prev: Link = None
    self: Link = None

    def __init__(self_, **kvargs):

        if "next" in kvargs and kvargs["next"] is not None:
            if (
                type(kvargs["next"]).__name__
                == AssignmentsLinks.__annotations__["next"]
            ):
                self_.next = kvargs["next"]
            else:
                self_.next = Link(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if (
                type(kvargs["prev"]).__name__
                == AssignmentsLinks.__annotations__["prev"]
            ):
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Link(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == AssignmentsLinks.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsMeta:
    """

    Attributes
    ----------
    count: int
      The total number of assignments matching the current filter.
    """

    count: int = None

    def __init__(self_, **kvargs):

        if "count" in kvargs and kvargs["count"] is not None:
            self_.count = kvargs["count"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Link:
    """

    Attributes
    ----------
    href: str
      URL that defines the resource.
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SpaceCreate:
    """

    Attributes
    ----------
    description: str
      The description of the space. Personal spaces do not have a description.
    name: str
      The name of the space. Personal spaces do not have a name.
    type: Literal["shared", "managed", "data"]
      The type of space such as shared, managed, and so on.
    """

    description: str = None
    name: str = None
    type: Literal["shared", "managed", "data"] = None

    def __init__(self_, **kvargs):

        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SpaceLinks:
    """

    Attributes
    ----------
    assignments: Link
    self: Link
    """

    assignments: Link = None
    self: Link = None

    def __init__(self_, **kvargs):

        if "assignments" in kvargs and kvargs["assignments"] is not None:
            if (
                type(kvargs["assignments"]).__name__
                == SpaceLinks.__annotations__["assignments"]
            ):
                self_.assignments = kvargs["assignments"]
            else:
                self_.assignments = Link(**kvargs["assignments"])
        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == SpaceLinks.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SpaceMeta:
    """
    Information about the space settings.

    Attributes
    ----------
    actions: list[str]
      The list of actions allowed by the current user in this space.
    assignableRoles: list[str]
      The list of roles that could be assigned in this space.
    roles: list[str]
      The list of roles assigned to the current user in this space.
    """

    actions: list[str] = None
    assignableRoles: list[str] = None
    roles: list[str] = None

    def __init__(self_, **kvargs):

        if "actions" in kvargs and kvargs["actions"] is not None:
            self_.actions = kvargs["actions"]
        if "assignableRoles" in kvargs and kvargs["assignableRoles"] is not None:
            self_.assignableRoles = kvargs["assignableRoles"]
        if "roles" in kvargs and kvargs["roles"] is not None:
            self_.roles = kvargs["roles"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SpacePatch:
    """

    Attributes
    ----------
    """

    def __init__(self_, **kvargs):
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SpaceTypes:
    """
    The distinct types of spaces (shared, managed, and so on).

    Attributes
    ----------
    data: list[str]
    """

    data: list[str] = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            self_.data = kvargs["data"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SpaceUpdate:
    """

    Attributes
    ----------
    description: str
      The description of the space. Personal spaces do not have a description.
    name: str
      The name of the space.
    ownerId: str
      The user ID of the space owner.
    """

    description: str = None
    name: str = None
    ownerId: str = None

    def __init__(self_, **kvargs):

        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "ownerId" in kvargs and kvargs["ownerId"] is not None:
            self_.ownerId = kvargs["ownerId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SpacesClass:
    """

    Attributes
    ----------
    data: list[Space]
    links: SpacesLinks
    meta: SpacesMeta
    """

    data: list[Space] = None
    links: SpacesLinks = None
    meta: SpacesMeta = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]" == SpacesClass.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [Space(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == SpacesClass.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = SpacesLinks(**kvargs["links"])
        if "meta" in kvargs and kvargs["meta"] is not None:
            if type(kvargs["meta"]).__name__ == SpacesClass.__annotations__["meta"]:
                self_.meta = kvargs["meta"]
            else:
                self_.meta = SpacesMeta(**kvargs["meta"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SpacesLinks:
    """

    Attributes
    ----------
    next: Link
    prev: Link
    self: Link
    """

    next: Link = None
    prev: Link = None
    self: Link = None

    def __init__(self_, **kvargs):

        if "next" in kvargs and kvargs["next"] is not None:
            if type(kvargs["next"]).__name__ == SpacesLinks.__annotations__["next"]:
                self_.next = kvargs["next"]
            else:
                self_.next = Link(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if type(kvargs["prev"]).__name__ == SpacesLinks.__annotations__["prev"]:
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Link(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == SpacesLinks.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SpacesMeta:
    """

    Attributes
    ----------
    count: int
      The total number of spaces matching the current filter.
    personalSpace: SpacesMetaPersonalSpace
      The meta related to personal space when applicable.
    """

    count: int = None
    personalSpace: SpacesMetaPersonalSpace = None

    def __init__(self_, **kvargs):

        if "count" in kvargs and kvargs["count"] is not None:
            self_.count = kvargs["count"]
        if "personalSpace" in kvargs and kvargs["personalSpace"] is not None:
            if (
                type(kvargs["personalSpace"]).__name__
                == SpacesMeta.__annotations__["personalSpace"]
            ):
                self_.personalSpace = kvargs["personalSpace"]
            else:
                self_.personalSpace = SpacesMetaPersonalSpace(**kvargs["personalSpace"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SpacesMetaPersonalSpace:
    """
    The meta related to personal space when applicable.

    Attributes
    ----------
    actions: list[str]
      The list of actions allowed by the current user in this space.
    resourceType: str
      resource type
    """

    actions: list[str] = None
    resourceType: str = None

    def __init__(self_, **kvargs):

        if "actions" in kvargs and kvargs["actions"] is not None:
            self_.actions = kvargs["actions"]
        if "resourceType" in kvargs and kvargs["resourceType"] is not None:
            self_.resourceType = kvargs["resourceType"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class Spaces:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def get_types(self) -> SpaceTypes:
        """
        Gets a list of distinct space types.

        Parameters
        ----------
        """
        response = self.auth.rest(
            path="/spaces/types",
            method="GET",
            params={},
            data=None,
        )
        obj = SpaceTypes(**response.json())
        obj.auth = self.auth
        return obj

    def get(self, spaceId: str) -> Space:
        """
        Retrieves a single space by ID.

        Parameters
        ----------
        spaceId: str
          The ID of the space to retrieve.
        """
        response = self.auth.rest(
            path="/spaces/{spaceId}".replace("{spaceId}", spaceId),
            method="GET",
            params={},
            data=None,
        )
        obj = Space(**response.json())
        obj.auth = self.auth
        return obj

    def get_spaces(
        self,
        action: str = None,
        filter: str = None,
        limit: int = 10,
        name: str = None,
        next: str = None,
        ownerId: str = None,
        prev: str = None,
        sort: str = None,
        type: str = None,
        max_items: int = 10,
    ) -> ListableResource[Space]:
        """
        Retrieves spaces that the current user has access to and match the query.

        Parameters
        ----------
        action: str = None
          Action on space. For example, "?action=publish".
        filter: str = None
          Exact match filtering on space name using SCIM. Case insensitive on attribute name. For example ?filter=name eq "MySpace" and ?filter=NAME eq "MySpace" is both valid.
        limit: int = 10
          Maximum number of spaces to return.
        name: str = None
          Space name to search and filter for. Case-insensitive open search with wildcards both as prefix and suffix. For example, "?name=fin" will get "finance", "Final" and "Griffin".
        next: str = None
          The next page cursor. Next links make use of this.
        ownerId: str = None
          Space ownerId to filter by. For example, "?ownerId=123".
        prev: str = None
          The previous page cursor. Previous links make use of this.
        sort: str = None
          Field to sort by. Prefix with +/- to indicate asc/desc. For example, "?sort=+name" to sort ascending on Name. Supported fields are "type", "name" and "createdAt".
        type: str = None
          Type(s) of space to filter. For example, "?type=managed,shared".
        """
        query_params = {}
        if action is not None:
            query_params["action"] = action
        if filter is not None:
            query_params["filter"] = filter

            warnings.warn("filter is experimental", UserWarning, stacklevel=2)
        if limit is not None:
            query_params["limit"] = limit
        if name is not None:
            query_params["name"] = name
        if next is not None:
            query_params["next"] = next
        if ownerId is not None:
            query_params["ownerId"] = ownerId
        if prev is not None:
            query_params["prev"] = prev
        if sort is not None:
            query_params["sort"] = sort
        if type is not None:
            query_params["type"] = type
        response = self.auth.rest(
            path="/spaces",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=Space,
            auth=self.auth,
            path="/spaces",
            max_items=max_items,
            query_params=query_params,
        )

    def create(self, data: SpaceCreate) -> Space:
        """
        Creates a space.

        Parameters
        ----------
        data: SpaceCreate
          Attributes that the user wants to set for a new space.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/spaces",
            method="POST",
            params={},
            data=data,
        )
        obj = Space(**response.json())
        obj.auth = self.auth
        return obj



from enum import Enum


class AuthType(Enum):
    """AuthType - represents a type of authentication"""

    APIKey = "APIKey"
    """ bearer authentication - using token representing a user in your tenant """
    OAuth2 = "OAuth2"
    """ OAuth2 authentication flows """



# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import List, Literal

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class Group:
    """
    represents a Group document

    Attributes
    ----------
    assignedRoles: list[GroupAssignedRoles]
    createdAt: str
      The timestamp for when the group record was created.
    id: str
      The unique identifier for the group
    idpId: str
      The unique identifier for the source IDP.
    lastUpdatedAt: str
      The timestamp for when the group record was last updated.
    links: GroupLinks
      Contains Links for current document
    name: str
      The name of the group.
    status: Literal["active", "disabled"]
      The state of the group.
    tenantId: str
      The tenant identifier associated with the given group
    """

    assignedRoles: list[GroupAssignedRoles] = None
    createdAt: str = None
    id: str = None
    idpId: str = None
    lastUpdatedAt: str = None
    links: GroupLinks = None
    name: str = None
    status: Literal["active", "disabled"] = None
    tenantId: str = None

    def __init__(self_, **kvargs):

        if "assignedRoles" in kvargs and kvargs["assignedRoles"] is not None:
            if all(
                f"list[{type(e).__name__}]" == Group.__annotations__["assignedRoles"]
                for e in kvargs["assignedRoles"]
            ):
                self_.assignedRoles = kvargs["assignedRoles"]
            else:
                self_.assignedRoles = [
                    GroupAssignedRoles(**e) for e in kvargs["assignedRoles"]
                ]
        if "createdAt" in kvargs and kvargs["createdAt"] is not None:
            self_.createdAt = kvargs["createdAt"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "idpId" in kvargs and kvargs["idpId"] is not None:
            self_.idpId = kvargs["idpId"]
        if "lastUpdatedAt" in kvargs and kvargs["lastUpdatedAt"] is not None:
            self_.lastUpdatedAt = kvargs["lastUpdatedAt"]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == Group.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = GroupLinks(**kvargs["links"])
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "status" in kvargs and kvargs["status"] is not None:
            self_.status = kvargs["status"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)

    def delete(self) -> None:
        """
        Delete group by id

        Parameters
        ----------
        """
        self.auth.rest(
            path="/groups/{groupId}".replace("{groupId}", self.id),
            method="DELETE",
            params={},
            data=None,
        )


@dataclass
class Filter:
    """
    An advanced query filter to be used for complex user querying in the tenant.

    Attributes
    ----------
    filter: str
      The advanced filtering to be applied the query. All conditional statements within this query parameter are case insensitive.
    """

    filter: str = None

    def __init__(self_, **kvargs):

        if "filter" in kvargs and kvargs["filter"] is not None:
            self_.filter = kvargs["filter"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GroupAssignedRoles:
    """
    represents a role entity to be stored on a Group entity, either default or custom role

    Attributes
    ----------
    id: str
    level: Literal["admin", "user"]
    name: str
    permissions: list[str]
    type: Literal["default"]
    """

    id: str = None
    level: Literal["admin", "user"] = None
    name: str = None
    permissions: list[str] = None
    type: Literal["default"] = None

    def __init__(self_, **kvargs):

        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "level" in kvargs and kvargs["level"] is not None:
            self_.level = kvargs["level"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "permissions" in kvargs and kvargs["permissions"] is not None:
            self_.permissions = kvargs["permissions"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GroupLinks:
    """
    Contains Links for current document

    Attributes
    ----------
    self: GroupLinksSelf
    """

    self: GroupLinksSelf = None

    def __init__(self_, **kvargs):

        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == GroupLinks.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = GroupLinksSelf(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GroupLinksSelf:
    """

    Attributes
    ----------
    href: str
      Link to the current group document
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GroupSettings:
    """
    represents a GroupSetting document

    Attributes
    ----------
    autoCreateGroups: bool
      Determines if groups should be created on login.
    links: GroupSettingsLinks
      Contains Links for current document
    syncIdpGroups: bool
      Determines if groups should be created on login.
    tenantId: str
      The unique tenant identifier.
    """

    autoCreateGroups: bool = None
    links: GroupSettingsLinks = None
    syncIdpGroups: bool = None
    tenantId: str = None

    def __init__(self_, **kvargs):

        if "autoCreateGroups" in kvargs and kvargs["autoCreateGroups"] is not None:
            self_.autoCreateGroups = kvargs["autoCreateGroups"]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == GroupSettings.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = GroupSettingsLinks(**kvargs["links"])
        if "syncIdpGroups" in kvargs and kvargs["syncIdpGroups"] is not None:
            self_.syncIdpGroups = kvargs["syncIdpGroups"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GroupSettingsLinks:
    """
    Contains Links for current document

    Attributes
    ----------
    self: GroupSettingsLinksSelf
    """

    self: GroupSettingsLinksSelf = None

    def __init__(self_, **kvargs):

        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == GroupSettingsLinks.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = GroupSettingsLinksSelf(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GroupSettingsLinksSelf:
    """

    Attributes
    ----------
    href: str
      Link to the current group settings document
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GroupsClass:
    """
    A result object when listing groups.

    Attributes
    ----------
    data: list[Group]
      An array of groups.
    links: GroupsLinks
    totalResults: int
      Indicates the total number of matching documents. Will only be returned if the query parameter "totalResults" is true.
    """

    data: list[Group] = None
    links: GroupsLinks = None
    totalResults: int = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]" == GroupsClass.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [Group(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == GroupsClass.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = GroupsLinks(**kvargs["links"])
        if "totalResults" in kvargs and kvargs["totalResults"] is not None:
            self_.totalResults = kvargs["totalResults"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GroupsLinks:
    """

    Attributes
    ----------
    next: GroupsLinksNext
    prev: GroupsLinksPrev
    self: GroupsLinksSelf
    """

    next: GroupsLinksNext = None
    prev: GroupsLinksPrev = None
    self: GroupsLinksSelf = None

    def __init__(self_, **kvargs):

        if "next" in kvargs and kvargs["next"] is not None:
            if type(kvargs["next"]).__name__ == GroupsLinks.__annotations__["next"]:
                self_.next = kvargs["next"]
            else:
                self_.next = GroupsLinksNext(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if type(kvargs["prev"]).__name__ == GroupsLinks.__annotations__["prev"]:
                self_.prev = kvargs["prev"]
            else:
                self_.prev = GroupsLinksPrev(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == GroupsLinks.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = GroupsLinksSelf(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GroupsLinksNext:
    """

    Attributes
    ----------
    href: str
      Link to the next page of items
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GroupsLinksPrev:
    """

    Attributes
    ----------
    href: str
      Link to the previous page of items
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GroupsLinksSelf:
    """

    Attributes
    ----------
    href: str
      Link to the current page of items
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SettingsPatch:
    """
    A JSON Patch document as defined in http://tools.ietf.org/html/rfc6902.

    Attributes
    ----------
    op: Literal["replace"]
      The operation to be performed.
    path: Literal["/autoCreateGroups", "/syncIdpGroups"]
      A JSON Pointer.
    value: bool
      The value to be used for this operation.
    """

    op: Literal["replace"] = None
    path: Literal["/autoCreateGroups", "/syncIdpGroups"] = None
    value: bool = None

    def __init__(self_, **kvargs):

        if "op" in kvargs and kvargs["op"] is not None:
            self_.op = kvargs["op"]
        if "path" in kvargs and kvargs["path"] is not None:
            self_.path = kvargs["path"]
        if "value" in kvargs and kvargs["value"] is not None:
            self_.value = kvargs["value"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SettingsPatchSchema(List["SettingsPatch"]):
    """
    An array of JSON Patches for the groups settings.

    Attributes
    ----------
    """

    def __init__(self_, elements):
        for e in elements:
            self_.append(SettingsPatch(**e))


class Groups:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def filter(
        self,
        data: Filter = None,
        limit: float = 20,
        next: str = None,
        prev: str = None,
        sort: Literal["name", "+name", "-name"] = "+name",
        max_items: int = 20,
    ) -> ListableResource[Group]:
        """
        Filter groups
        Retrieves a list of groups matching the filter using advanced query string.

        Parameters
        ----------
        limit: float = 20
          The number of user entries to retrieve.
        next: str = None
          Get users with IDs that are higher than the target user ID. Cannot be used in conjunction with prev.
        prev: str = None
          Get users with IDs that are lower than the target user ID. Cannot be used in conjunction with next.
        sort: Literal["name", "+name", "-name"] = "+name"
          The field to sort by, with +/- prefix indicating sort order
        data: Filter = None
          Will contain the query filter to apply. It shall not contain more than 50 ids.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        query_params = {}
        if limit is not None:
            query_params["limit"] = limit
        if next is not None:
            query_params["next"] = next
        if prev is not None:
            query_params["prev"] = prev
        if sort is not None:
            query_params["sort"] = sort
        response = self.auth.rest(
            path="/groups/actions/filter",
            method="POST",
            params=query_params,
            data=data,
        )
        return ListableResource(
            response=response.json(),
            cls=Group,
            auth=self.auth,
            path="/groups/actions/filter",
            max_items=max_items,
            query_params=query_params,
        )

    def get_settings(self) -> GroupSettings:
        """
        Get group settings
        Returns the active tenant's group settings.

        Parameters
        ----------
        """
        response = self.auth.rest(
            path="/groups/settings",
            method="GET",
            params={},
            data=None,
        )
        obj = GroupSettings(**response.json())
        obj.auth = self.auth
        return obj

    def patch_settings(self, data: SettingsPatchSchema) -> None:
        """
        Update group settings

        Parameters
        ----------
        data: SettingsPatchSchema

        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        self.auth.rest(
            path="/groups/settings",
            method="PATCH",
            params={},
            data=data,
        )

    def get(self, groupId: str) -> Group:
        """
        Get group by id
        Returns the requested group.

        Parameters
        ----------
        groupId: str
          The group's unique identifier
        """
        response = self.auth.rest(
            path="/groups/{groupId}".replace("{groupId}", groupId),
            method="GET",
            params={},
            data=None,
        )
        obj = Group(**response.json())
        obj.auth = self.auth
        return obj

    def get_groups(
        self,
        filter: str = None,
        limit: float = 20,
        next: str = None,
        prev: str = None,
        sort: str = None,
        totalResults: bool = None,
        max_items: int = 20,
    ) -> ListableResource[Group]:
        """
        List groups.
        Returns a list of groups with cursor-based pagination.

        Parameters
        ----------
        filter: str = None
          The advanced filtering to use for the query. Refer to RFC 7644 https://datatracker.ietf.org/doc/rfc7644/ for the syntax. Cannot be combined with any of the fields marked as deprecated. All conditional statements within this query parameter are case insensitive.
        limit: float = 20
          The number of groups to retrieve.
        next: str = None
          The next page cursor.
        prev: str = None
          The previous page cursor.
        sort: str = None
          Optional resource field name to sort on, eg. name. Can be prefixed with +/- to determine order, defaults to (+) ascending.
        totalResults: bool = None
          Whether to return a total match count in the result. Defaults to false.
        """
        query_params = {}
        if filter is not None:
            query_params["filter"] = filter
        if limit is not None:
            query_params["limit"] = limit
        if next is not None:
            query_params["next"] = next
        if prev is not None:
            query_params["prev"] = prev
        if sort is not None:
            query_params["sort"] = sort
        if totalResults is not None:
            query_params["totalResults"] = totalResults
        response = self.auth.rest(
            path="/groups",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=Group,
            auth=self.auth,
            path="/groups",
            max_items=max_items,
            query_params=query_params,
        )



# This is spectacularly generated code by spectacular v0.0.0 based on
# Qlik Cloud Services 0.384.10

from __future__ import annotations

from dataclasses import asdict, dataclass

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class AssignmentsActionsAddRequest:
    """

    Attributes
    ----------
    add: list[AssignmentsActionsAddRequestAdd]
    """

    add: list[AssignmentsActionsAddRequestAdd] = None

    def __init__(self_, **kvargs):
        if "add" in kvargs:
            if type(kvargs["add"]).__name__ == self_.__annotations__["add"]:
                self_.add = kvargs["add"]
            else:
                self_.add = [
                    AssignmentsActionsAddRequestAdd(**e) for e in kvargs["add"]
                ]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsActionsAddRequestAdd:
    """

    Attributes
    ----------
    name: str
      User name
    subject: str
      User subject
    type: str
      Allotment type
    userId: str
      User ID
    """

    name: str = None
    subject: str = None
    type: str = None
    userId: str = None

    def __init__(self_, **kvargs):
        if "name" in kvargs:
            if type(kvargs["name"]).__name__ == self_.__annotations__["name"]:
                self_.name = kvargs["name"]
            else:
                self_.name = kvargs["name"]
        if "subject" in kvargs:
            if type(kvargs["subject"]).__name__ == self_.__annotations__["subject"]:
                self_.subject = kvargs["subject"]
            else:
                self_.subject = kvargs["subject"]
        if "type" in kvargs:
            if type(kvargs["type"]).__name__ == self_.__annotations__["type"]:
                self_.type = kvargs["type"]
            else:
                self_.type = kvargs["type"]
        if "userId" in kvargs:
            if type(kvargs["userId"]).__name__ == self_.__annotations__["userId"]:
                self_.userId = kvargs["userId"]
            else:
                self_.userId = kvargs["userId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsActionsAddResponse:
    """

    Attributes
    ----------
    data: list[AssignmentsActionsAddResponseData]
    """

    data: list[AssignmentsActionsAddResponseData] = None

    def __init__(self_, **kvargs):
        if "data" in kvargs:
            if type(kvargs["data"]).__name__ == self_.__annotations__["data"]:
                self_.data = kvargs["data"]
            else:
                self_.data = [
                    AssignmentsActionsAddResponseData(**e) for e in kvargs["data"]
                ]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsActionsAddResponseData:
    """

    Attributes
    ----------
    code: str
      Error code
    status: int
      Response status
    subject: str
      Subject
    title: str
      Error title
    type: str
      Allotment type
    """

    code: str = None
    status: int = None
    subject: str = None
    title: str = None
    type: str = None

    def __init__(self_, **kvargs):
        if "code" in kvargs:
            if type(kvargs["code"]).__name__ == self_.__annotations__["code"]:
                self_.code = kvargs["code"]
            else:
                self_.code = kvargs["code"]
        if "status" in kvargs:
            if type(kvargs["status"]).__name__ == self_.__annotations__["status"]:
                self_.status = kvargs["status"]
            else:
                self_.status = kvargs["status"]
        if "subject" in kvargs:
            if type(kvargs["subject"]).__name__ == self_.__annotations__["subject"]:
                self_.subject = kvargs["subject"]
            else:
                self_.subject = kvargs["subject"]
        if "title" in kvargs:
            if type(kvargs["title"]).__name__ == self_.__annotations__["title"]:
                self_.title = kvargs["title"]
            else:
                self_.title = kvargs["title"]
        if "type" in kvargs:
            if type(kvargs["type"]).__name__ == self_.__annotations__["type"]:
                self_.type = kvargs["type"]
            else:
                self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsActionsDeleteRequest:
    """

    Attributes
    ----------
    delete: list[AssignmentsActionsDeleteRequestDelete]
    """

    delete: list[AssignmentsActionsDeleteRequestDelete] = None

    def __init__(self_, **kvargs):
        if "delete" in kvargs:
            if type(kvargs["delete"]).__name__ == self_.__annotations__["delete"]:
                self_.delete = kvargs["delete"]
            else:
                self_.delete = [
                    AssignmentsActionsDeleteRequestDelete(**e) for e in kvargs["delete"]
                ]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsActionsDeleteRequestDelete:
    """

    Attributes
    ----------
    subject: str
      User subject
    type: str
      Allotment type
    """

    subject: str = None
    type: str = None

    def __init__(self_, **kvargs):
        if "subject" in kvargs:
            if type(kvargs["subject"]).__name__ == self_.__annotations__["subject"]:
                self_.subject = kvargs["subject"]
            else:
                self_.subject = kvargs["subject"]
        if "type" in kvargs:
            if type(kvargs["type"]).__name__ == self_.__annotations__["type"]:
                self_.type = kvargs["type"]
            else:
                self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsActionsDeleteResponse:
    """

    Attributes
    ----------
    data: list[AssignmentsActionsDeleteResponseData]
    """

    data: list[AssignmentsActionsDeleteResponseData] = None

    def __init__(self_, **kvargs):
        if "data" in kvargs:
            if type(kvargs["data"]).__name__ == self_.__annotations__["data"]:
                self_.data = kvargs["data"]
            else:
                self_.data = [
                    AssignmentsActionsDeleteResponseData(**e) for e in kvargs["data"]
                ]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsActionsDeleteResponseData:
    """

    Attributes
    ----------
    code: str
      Error code
    status: int
      Response status
    subject: str
      Subject
    title: str
      Error title
    type: str
      Allotment type
    """

    code: str = None
    status: int = None
    subject: str = None
    title: str = None
    type: str = None

    def __init__(self_, **kvargs):
        if "code" in kvargs:
            if type(kvargs["code"]).__name__ == self_.__annotations__["code"]:
                self_.code = kvargs["code"]
            else:
                self_.code = kvargs["code"]
        if "status" in kvargs:
            if type(kvargs["status"]).__name__ == self_.__annotations__["status"]:
                self_.status = kvargs["status"]
            else:
                self_.status = kvargs["status"]
        if "subject" in kvargs:
            if type(kvargs["subject"]).__name__ == self_.__annotations__["subject"]:
                self_.subject = kvargs["subject"]
            else:
                self_.subject = kvargs["subject"]
        if "title" in kvargs:
            if type(kvargs["title"]).__name__ == self_.__annotations__["title"]:
                self_.title = kvargs["title"]
            else:
                self_.title = kvargs["title"]
        if "type" in kvargs:
            if type(kvargs["type"]).__name__ == self_.__annotations__["type"]:
                self_.type = kvargs["type"]
            else:
                self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsActionsUpdateRequest:
    """

    Attributes
    ----------
    update: list[AssignmentsActionsUpdateRequestUpdate]
    """

    update: list[AssignmentsActionsUpdateRequestUpdate] = None

    def __init__(self_, **kvargs):
        if "update" in kvargs:
            if type(kvargs["update"]).__name__ == self_.__annotations__["update"]:
                self_.update = kvargs["update"]
            else:
                self_.update = [
                    AssignmentsActionsUpdateRequestUpdate(**e) for e in kvargs["update"]
                ]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsActionsUpdateRequestUpdate:
    """

    Attributes
    ----------
    sourceType: str
      Current assignment type.
    subject: str
      User subject
    type: str
      Target assignment type.
    """

    sourceType: str = None
    subject: str = None
    type: str = None

    def __init__(self_, **kvargs):
        if "sourceType" in kvargs:
            if (
                type(kvargs["sourceType"]).__name__
                == self_.__annotations__["sourceType"]
            ):
                self_.sourceType = kvargs["sourceType"]
            else:
                self_.sourceType = kvargs["sourceType"]
        if "subject" in kvargs:
            if type(kvargs["subject"]).__name__ == self_.__annotations__["subject"]:
                self_.subject = kvargs["subject"]
            else:
                self_.subject = kvargs["subject"]
        if "type" in kvargs:
            if type(kvargs["type"]).__name__ == self_.__annotations__["type"]:
                self_.type = kvargs["type"]
            else:
                self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsActionsUpdateResponse:
    """

    Attributes
    ----------
    data: list[AssignmentsActionsUpdateResponseData]
    """

    data: list[AssignmentsActionsUpdateResponseData] = None

    def __init__(self_, **kvargs):
        if "data" in kvargs:
            if type(kvargs["data"]).__name__ == self_.__annotations__["data"]:
                self_.data = kvargs["data"]
            else:
                self_.data = [
                    AssignmentsActionsUpdateResponseData(**e) for e in kvargs["data"]
                ]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsActionsUpdateResponseData:
    """

    Attributes
    ----------
    code: str
      Error code
    sourceType: str
      Current allotment type.
    status: int
      Response status
    subject: str
      Subject
    title: str
      Error title
    type: str
      Target allotment type.
    """

    code: str = None
    sourceType: str = None
    status: int = None
    subject: str = None
    title: str = None
    type: str = None

    def __init__(self_, **kvargs):
        if "code" in kvargs:
            if type(kvargs["code"]).__name__ == self_.__annotations__["code"]:
                self_.code = kvargs["code"]
            else:
                self_.code = kvargs["code"]
        if "sourceType" in kvargs:
            if (
                type(kvargs["sourceType"]).__name__
                == self_.__annotations__["sourceType"]
            ):
                self_.sourceType = kvargs["sourceType"]
            else:
                self_.sourceType = kvargs["sourceType"]
        if "status" in kvargs:
            if type(kvargs["status"]).__name__ == self_.__annotations__["status"]:
                self_.status = kvargs["status"]
            else:
                self_.status = kvargs["status"]
        if "subject" in kvargs:
            if type(kvargs["subject"]).__name__ == self_.__annotations__["subject"]:
                self_.subject = kvargs["subject"]
            else:
                self_.subject = kvargs["subject"]
        if "title" in kvargs:
            if type(kvargs["title"]).__name__ == self_.__annotations__["title"]:
                self_.title = kvargs["title"]
            else:
                self_.title = kvargs["title"]
        if "type" in kvargs:
            if type(kvargs["type"]).__name__ == self_.__annotations__["type"]:
                self_.type = kvargs["type"]
            else:
                self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsResponse:
    """

    Attributes
    ----------
    data: list[AssignmentsResponseData]
    links: AssignmentsResponseLinks
    """

    data: list[AssignmentsResponseData] = None
    links: AssignmentsResponseLinks = None

    def __init__(self_, **kvargs):
        if "data" in kvargs:
            if type(kvargs["data"]).__name__ == self_.__annotations__["data"]:
                self_.data = kvargs["data"]
            else:
                self_.data = [AssignmentsResponseData(**e) for e in kvargs["data"]]
        if "links" in kvargs:
            if type(kvargs["links"]).__name__ == self_.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = AssignmentsResponseLinks(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsResponseData:
    """

    Attributes
    ----------
    created: str
      Assignment created date.
    excess: bool
      Assignment excess status.
    name: str
      User name
    subject: str
      Subject
    type: str
      Allotment type
    userId: str
      User ID
    """

    created: str = None
    excess: bool = None
    name: str = None
    subject: str = None
    type: str = None
    userId: str = None

    def __init__(self_, **kvargs):
        if "created" in kvargs:
            if type(kvargs["created"]).__name__ == self_.__annotations__["created"]:
                self_.created = kvargs["created"]
            else:
                self_.created = kvargs["created"]
        if "excess" in kvargs:
            if type(kvargs["excess"]).__name__ == self_.__annotations__["excess"]:
                self_.excess = kvargs["excess"]
            else:
                self_.excess = kvargs["excess"]
        if "name" in kvargs:
            if type(kvargs["name"]).__name__ == self_.__annotations__["name"]:
                self_.name = kvargs["name"]
            else:
                self_.name = kvargs["name"]
        if "subject" in kvargs:
            if type(kvargs["subject"]).__name__ == self_.__annotations__["subject"]:
                self_.subject = kvargs["subject"]
            else:
                self_.subject = kvargs["subject"]
        if "type" in kvargs:
            if type(kvargs["type"]).__name__ == self_.__annotations__["type"]:
                self_.type = kvargs["type"]
            else:
                self_.type = kvargs["type"]
        if "userId" in kvargs:
            if type(kvargs["userId"]).__name__ == self_.__annotations__["userId"]:
                self_.userId = kvargs["userId"]
            else:
                self_.userId = kvargs["userId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class AssignmentsResponseLinks:
    """

    Attributes
    ----------
    next: Licenseshref
    prev: Licenseshref
    """

    next: Licenseshref = None
    prev: Licenseshref = None

    def __init__(self_, **kvargs):
        if "next" in kvargs:
            if type(kvargs["next"]).__name__ == self_.__annotations__["next"]:
                self_.next = kvargs["next"]
            else:
                self_.next = Licenseshref(**kvargs["next"])
        if "prev" in kvargs:
            if type(kvargs["prev"]).__name__ == self_.__annotations__["prev"]:
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Licenseshref(**kvargs["prev"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ConsumptionEventsResponse:
    """

    Attributes
    ----------
    data: list[ConsumptionEventsResponseData]
    links: ConsumptionEventsResponseLinks
    """

    data: list[ConsumptionEventsResponseData] = None
    links: ConsumptionEventsResponseLinks = None

    def __init__(self_, **kvargs):
        if "data" in kvargs:
            if type(kvargs["data"]).__name__ == self_.__annotations__["data"]:
                self_.data = kvargs["data"]
            else:
                self_.data = [
                    ConsumptionEventsResponseData(**e) for e in kvargs["data"]
                ]
        if "links" in kvargs:
            if type(kvargs["links"]).__name__ == self_.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = ConsumptionEventsResponseLinks(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ConsumptionEventsResponseData:
    """

    Attributes
    ----------
    allotmentId: str
      Allotment ID
    appId: str
      App ID
    capacityUsed: int
      Analyzer capacity chunks consumed.
    duration: str
      Engine session duration.
    endTime: str
      Engine session end time.
    id: str
      ID
    licenseUsage: str
      License usage
    minutesUsed: int
      Analyzer capacity minutes consumed.
    sessionId: str
      Engine session ID.
    userId: str
      User ID
    """

    allotmentId: str = None
    appId: str = None
    capacityUsed: int = None
    duration: str = None
    endTime: str = None
    id: str = None
    licenseUsage: str = None
    minutesUsed: int = None
    sessionId: str = None
    userId: str = None

    def __init__(self_, **kvargs):
        if "allotmentId" in kvargs:
            if (
                type(kvargs["allotmentId"]).__name__
                == self_.__annotations__["allotmentId"]
            ):
                self_.allotmentId = kvargs["allotmentId"]
            else:
                self_.allotmentId = kvargs["allotmentId"]
        if "appId" in kvargs:
            if type(kvargs["appId"]).__name__ == self_.__annotations__["appId"]:
                self_.appId = kvargs["appId"]
            else:
                self_.appId = kvargs["appId"]
        if "capacityUsed" in kvargs:
            if (
                type(kvargs["capacityUsed"]).__name__
                == self_.__annotations__["capacityUsed"]
            ):
                self_.capacityUsed = kvargs["capacityUsed"]
            else:
                self_.capacityUsed = kvargs["capacityUsed"]
        if "duration" in kvargs:
            if type(kvargs["duration"]).__name__ == self_.__annotations__["duration"]:
                self_.duration = kvargs["duration"]
            else:
                self_.duration = kvargs["duration"]
        if "endTime" in kvargs:
            if type(kvargs["endTime"]).__name__ == self_.__annotations__["endTime"]:
                self_.endTime = kvargs["endTime"]
            else:
                self_.endTime = kvargs["endTime"]
        if "id" in kvargs:
            if type(kvargs["id"]).__name__ == self_.__annotations__["id"]:
                self_.id = kvargs["id"]
            else:
                self_.id = kvargs["id"]
        if "licenseUsage" in kvargs:
            if (
                type(kvargs["licenseUsage"]).__name__
                == self_.__annotations__["licenseUsage"]
            ):
                self_.licenseUsage = kvargs["licenseUsage"]
            else:
                self_.licenseUsage = kvargs["licenseUsage"]
        if "minutesUsed" in kvargs:
            if (
                type(kvargs["minutesUsed"]).__name__
                == self_.__annotations__["minutesUsed"]
            ):
                self_.minutesUsed = kvargs["minutesUsed"]
            else:
                self_.minutesUsed = kvargs["minutesUsed"]
        if "sessionId" in kvargs:
            if type(kvargs["sessionId"]).__name__ == self_.__annotations__["sessionId"]:
                self_.sessionId = kvargs["sessionId"]
            else:
                self_.sessionId = kvargs["sessionId"]
        if "userId" in kvargs:
            if type(kvargs["userId"]).__name__ == self_.__annotations__["userId"]:
                self_.userId = kvargs["userId"]
            else:
                self_.userId = kvargs["userId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ConsumptionEventsResponseLinks:
    """

    Attributes
    ----------
    next: Licenseshref
    prev: Licenseshref
    """

    next: Licenseshref = None
    prev: Licenseshref = None

    def __init__(self_, **kvargs):
        if "next" in kvargs:
            if type(kvargs["next"]).__name__ == self_.__annotations__["next"]:
                self_.next = kvargs["next"]
            else:
                self_.next = Licenseshref(**kvargs["next"])
        if "prev" in kvargs:
            if type(kvargs["prev"]).__name__ == self_.__annotations__["prev"]:
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Licenseshref(**kvargs["prev"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class LicenseOverview:
    """

    Attributes
    ----------
    allotments: list[LicenseOverviewAllotments]
    licenseKey: str
    licenseNumber: str
    origin: str
      Origin of license key.
    parameters: list[LicenseOverviewParameters]
      The license parameters.
    product: str
      The product the license is valid for.
    status: str
      Enum with status of license. Only status Ok grants license. access.
    trial: bool
      Boolean indicating if it is a trial license.
    updated: str
      An ISO 8601 timestamp for when the license was last updated.
    valid: str
      Period that the license is currently set to be active. Represented as an ISO 8601 time interval with start and end.
    """

    allotments: list[LicenseOverviewAllotments] = None
    licenseKey: str = None
    licenseNumber: str = None
    origin: str = None
    parameters: list[LicenseOverviewParameters] = None
    product: str = None
    status: str = None
    trial: bool = None
    updated: str = None
    valid: str = None

    def __init__(self_, **kvargs):
        if "allotments" in kvargs:
            if (
                type(kvargs["allotments"]).__name__
                == self_.__annotations__["allotments"]
            ):
                self_.allotments = kvargs["allotments"]
            else:
                self_.allotments = [
                    LicenseOverviewAllotments(**e) for e in kvargs["allotments"]
                ]
        if "licenseKey" in kvargs:
            if (
                type(kvargs["licenseKey"]).__name__
                == self_.__annotations__["licenseKey"]
            ):
                self_.licenseKey = kvargs["licenseKey"]
            else:
                self_.licenseKey = kvargs["licenseKey"]
        if "licenseNumber" in kvargs:
            if (
                type(kvargs["licenseNumber"]).__name__
                == self_.__annotations__["licenseNumber"]
            ):
                self_.licenseNumber = kvargs["licenseNumber"]
            else:
                self_.licenseNumber = kvargs["licenseNumber"]
        if "origin" in kvargs:
            if type(kvargs["origin"]).__name__ == self_.__annotations__["origin"]:
                self_.origin = kvargs["origin"]
            else:
                self_.origin = kvargs["origin"]
        if "parameters" in kvargs:
            if (
                type(kvargs["parameters"]).__name__
                == self_.__annotations__["parameters"]
            ):
                self_.parameters = kvargs["parameters"]
            else:
                self_.parameters = [
                    LicenseOverviewParameters(**e) for e in kvargs["parameters"]
                ]
        if "product" in kvargs:
            if type(kvargs["product"]).__name__ == self_.__annotations__["product"]:
                self_.product = kvargs["product"]
            else:
                self_.product = kvargs["product"]
        if "status" in kvargs:
            if type(kvargs["status"]).__name__ == self_.__annotations__["status"]:
                self_.status = kvargs["status"]
            else:
                self_.status = kvargs["status"]
        if "trial" in kvargs:
            if type(kvargs["trial"]).__name__ == self_.__annotations__["trial"]:
                self_.trial = kvargs["trial"]
            else:
                self_.trial = kvargs["trial"]
        if "updated" in kvargs:
            if type(kvargs["updated"]).__name__ == self_.__annotations__["updated"]:
                self_.updated = kvargs["updated"]
            else:
                self_.updated = kvargs["updated"]
        if "valid" in kvargs:
            if type(kvargs["valid"]).__name__ == self_.__annotations__["valid"]:
                self_.valid = kvargs["valid"]
            else:
                self_.valid = kvargs["valid"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class LicenseOverviewAllotments:
    """

    Attributes
    ----------
    name: str
    overage: int
      Overage value; -1 means unbounded overage.
    units: int
    unitsUsed: int
    usageClass: str
    """

    name: str = None
    overage: int = None
    units: int = None
    unitsUsed: int = None
    usageClass: str = None

    def __init__(self_, **kvargs):
        if "name" in kvargs:
            if type(kvargs["name"]).__name__ == self_.__annotations__["name"]:
                self_.name = kvargs["name"]
            else:
                self_.name = kvargs["name"]
        if "overage" in kvargs:
            if type(kvargs["overage"]).__name__ == self_.__annotations__["overage"]:
                self_.overage = kvargs["overage"]
            else:
                self_.overage = kvargs["overage"]
        if "units" in kvargs:
            if type(kvargs["units"]).__name__ == self_.__annotations__["units"]:
                self_.units = kvargs["units"]
            else:
                self_.units = kvargs["units"]
        if "unitsUsed" in kvargs:
            if type(kvargs["unitsUsed"]).__name__ == self_.__annotations__["unitsUsed"]:
                self_.unitsUsed = kvargs["unitsUsed"]
            else:
                self_.unitsUsed = kvargs["unitsUsed"]
        if "usageClass" in kvargs:
            if (
                type(kvargs["usageClass"]).__name__
                == self_.__annotations__["usageClass"]
            ):
                self_.usageClass = kvargs["usageClass"]
            else:
                self_.usageClass = kvargs["usageClass"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class LicenseOverviewParameters:
    """

    Attributes
    ----------
    access: LicenseOverviewParametersAccess
      Parameters for licenses to control access to the parameters.
    name: str
      Parameter set (provision) name.
    valid: str
      Time interval for parameter validity.
    values: LicenseOverviewParametersValues
      Parameter values
    """

    access: LicenseOverviewParametersAccess = None
    name: str = None
    valid: str = None
    values: LicenseOverviewParametersValues = None

    def __init__(self_, **kvargs):
        if "access" in kvargs:
            if type(kvargs["access"]).__name__ == self_.__annotations__["access"]:
                self_.access = kvargs["access"]
            else:
                self_.access = LicenseOverviewParametersAccess(**kvargs["access"])
        if "name" in kvargs:
            if type(kvargs["name"]).__name__ == self_.__annotations__["name"]:
                self_.name = kvargs["name"]
            else:
                self_.name = kvargs["name"]
        if "valid" in kvargs:
            if type(kvargs["valid"]).__name__ == self_.__annotations__["valid"]:
                self_.valid = kvargs["valid"]
            else:
                self_.valid = kvargs["valid"]
        if "values" in kvargs:
            if type(kvargs["values"]).__name__ == self_.__annotations__["values"]:
                self_.values = kvargs["values"]
            else:
                self_.values = LicenseOverviewParametersValues(**kvargs["values"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class LicenseOverviewParametersAccess:
    """
    Parameters for licenses to control access to the parameters.

    Attributes
    ----------
    allotment: str
      Name of an allotment that the user must have access to. to
    """

    allotment: str = None

    def __init__(self_, **kvargs):
        if "allotment" in kvargs:
            if type(kvargs["allotment"]).__name__ == self_.__annotations__["allotment"]:
                self_.allotment = kvargs["allotment"]
            else:
                self_.allotment = kvargs["allotment"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class LicenseOverviewParametersValues:
    """
    Parameter values

    Attributes
    ----------
    """

    def __init__(self_, **kvargs):
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class LicenseStatus:
    """

    Attributes
    ----------
    origin: str
      Origin of license key.
    product: str
      The product the license is valid for.
    status: str
      Enum with status of license. Only status Ok grants license. access.
    trial: bool
      Boolean indicating if it is a trial license.
    type: str
      Type of license key.
    valid: str
      Period that the license is currently set to be active. Represented as an ISO 8601 time interval with start and end.
    """

    origin: str = None
    product: str = None
    status: str = None
    trial: bool = None
    type: str = None
    valid: str = None

    def __init__(self_, **kvargs):
        if "origin" in kvargs:
            if type(kvargs["origin"]).__name__ == self_.__annotations__["origin"]:
                self_.origin = kvargs["origin"]
            else:
                self_.origin = kvargs["origin"]
        if "product" in kvargs:
            if type(kvargs["product"]).__name__ == self_.__annotations__["product"]:
                self_.product = kvargs["product"]
            else:
                self_.product = kvargs["product"]
        if "status" in kvargs:
            if type(kvargs["status"]).__name__ == self_.__annotations__["status"]:
                self_.status = kvargs["status"]
            else:
                self_.status = kvargs["status"]
        if "trial" in kvargs:
            if type(kvargs["trial"]).__name__ == self_.__annotations__["trial"]:
                self_.trial = kvargs["trial"]
            else:
                self_.trial = kvargs["trial"]
        if "type" in kvargs:
            if type(kvargs["type"]).__name__ == self_.__annotations__["type"]:
                self_.type = kvargs["type"]
            else:
                self_.type = kvargs["type"]
        if "valid" in kvargs:
            if type(kvargs["valid"]).__name__ == self_.__annotations__["valid"]:
                self_.valid = kvargs["valid"]
            else:
                self_.valid = kvargs["valid"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Licenseshref:
    """

    Attributes
    ----------
    href: str
      link
    """

    href: str = None

    def __init__(self_, **kvargs):
        if "href" in kvargs:
            if type(kvargs["href"]).__name__ == self_.__annotations__["href"]:
                self_.href = kvargs["href"]
            else:
                self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SettingsBody:
    """

    Attributes
    ----------
    autoAssignAnalyzer: bool
      If analyzer users are available, they will be automatically assigned. Otherwise, analyzer capacity will be assigned, if available.
    autoAssignProfessional: bool
      If professional users are available, they will be automatically assigned. Otherwise, analyzer capacity will be assigned, if available.
    """

    autoAssignAnalyzer: bool = None
    autoAssignProfessional: bool = None

    def __init__(self_, **kvargs):
        if "autoAssignAnalyzer" in kvargs:
            if (
                type(kvargs["autoAssignAnalyzer"]).__name__
                == self_.__annotations__["autoAssignAnalyzer"]
            ):
                self_.autoAssignAnalyzer = kvargs["autoAssignAnalyzer"]
            else:
                self_.autoAssignAnalyzer = kvargs["autoAssignAnalyzer"]
        if "autoAssignProfessional" in kvargs:
            if (
                type(kvargs["autoAssignProfessional"]).__name__
                == self_.__annotations__["autoAssignProfessional"]
            ):
                self_.autoAssignProfessional = kvargs["autoAssignProfessional"]
            else:
                self_.autoAssignProfessional = kvargs["autoAssignProfessional"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class Licenses:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def add(self, data: AssignmentsActionsAddRequest) -> AssignmentsActionsAddResponse:
        """
        Assigns license access to the given users

        Parameters
        ----------
        authorization: str = None
          Authentication token
        data: AssignmentsActionsAddRequest
          List of subjects to allocate assignments for.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/licenses/assignments/actions/add",
            method="POST",
            params={},
            data=data,
        )
        obj = AssignmentsActionsAddResponse(**response.json())
        obj.auth = self.auth
        return obj

    def delete(
        self, data: AssignmentsActionsDeleteRequest
    ) -> AssignmentsActionsDeleteResponse:
        """
        Removes license access for the given users

        Parameters
        ----------
        authorization: str = None
          Authentication token
        data: AssignmentsActionsDeleteRequest
          List of assignments to delete.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/licenses/assignments/actions/delete",
            method="POST",
            params={},
            data=data,
        )
        obj = AssignmentsActionsDeleteResponse(**response.json())
        obj.auth = self.auth
        return obj

    def update(
        self, data: AssignmentsActionsUpdateRequest
    ) -> AssignmentsActionsUpdateResponse:
        """
        Updates license access for the given users

        Parameters
        ----------
        authorization: str = None
          Authentication token
        data: AssignmentsActionsUpdateRequest
          List of assignments to update.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/licenses/assignments/actions/update",
            method="POST",
            params={},
            data=data,
        )
        obj = AssignmentsActionsUpdateResponse(**response.json())
        obj.auth = self.auth
        return obj

    def get_assignments(
        self,
        filter: str = None,
        limit: int = 20,
        page: str = None,
        sort: str = None,
        max_items: int = 20,
    ) -> ListableResource[AssignmentsResponseData]:
        """
        Retrieves assignments for the current tenant

        Parameters
        ----------
        authorization: str = None
          Authentication token
        filter: str = None
          The filter for finding entries.
        limit: int = 20
          The preferred number of entries to return.
        page: str = None
          The requested page.
        sort: str = None
          The field to sort on; can be prefixed with +/- for ascending/descending sort order.
        """
        query_params = {}
        if filter is not None:
            query_params["filter"] = filter
        if limit is not None:
            query_params["limit"] = limit
        if page is not None:
            query_params["page"] = page
        if sort is not None:
            query_params["sort"] = sort
        response = self.auth.rest(
            path="/licenses/assignments",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=AssignmentsResponseData,
            auth=self.auth,
            path="/licenses/assignments",
            max_items=max_items,
            query_params=query_params,
        )

    def get_consumptions(
        self,
        filter: str = None,
        limit: int = 200,
        page: str = None,
        sort: str = None,
        max_items: int = 200,
    ) -> ListableResource[ConsumptionEventsResponseData]:
        """
        Retrieves license consumption for the current tenant

        Parameters
        ----------
        authorization: str = None
          Authentication token
        filter: str = None
          The filter for finding entries.
        limit: int = 200
          The preferred number of entries to return.
        page: str = None
          The requested page.
        sort: str = None
          The field to sort on; can be prefixed with +/- for ascending/descending sort order.
        """
        query_params = {}
        if filter is not None:
            query_params["filter"] = filter
        if limit is not None:
            query_params["limit"] = limit
        if page is not None:
            query_params["page"] = page
        if sort is not None:
            query_params["sort"] = sort
        response = self.auth.rest(
            path="/licenses/consumption",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=ConsumptionEventsResponseData,
            auth=self.auth,
            path="/licenses/consumption",
            max_items=max_items,
            query_params=query_params,
        )

    def get_overview(self) -> LicenseOverview:
        """
        Gets the general information of the license applied to the current tenant

        Parameters
        ----------
        authorization: str = None
          Authentication token. In the case of a user token, the tenant admin role is required.
        """
        response = self.auth.rest(
            path="/licenses/overview",
            method="GET",
            params={},
            data=None,
        )
        obj = LicenseOverview(**response.json())
        obj.auth = self.auth
        return obj

    def get_settings(self) -> SettingsBody:
        """
        Get auto assign settings for tenant.

        Parameters
        ----------
        authorization: str = None
          Authentication token
        """
        response = self.auth.rest(
            path="/licenses/settings",
            method="GET",
            params={},
            data=None,
        )
        obj = SettingsBody(**response.json())
        obj.auth = self.auth
        return obj

    def set_settings(self, data: SettingsBody = None) -> SettingsBody:
        """
        Set auto assign settings for tenant

        Parameters
        ----------
        authorization: str = None
          Authentication token
        data: SettingsBody = None
          Dynamic assignment settings for professional and analyzer users. If professional users and analyzer users are both set, professional users will be automatically assigned, if available. Otherwise, analyzer users will be assigned. If neither of those users are available, analyzer capacity will be assigned, if available.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/licenses/settings",
            method="PUT",
            params={},
            data=data,
        )
        obj = SettingsBody(**response.json())
        obj.auth = self.auth
        return obj

    def get_status(self) -> LicenseStatus:
        """
        Gets the license status information of the current tenant

        Parameters
        ----------
        authorization: str = None
          Authentication token
        """
        response = self.auth.rest(
            path="/licenses/status",
            method="GET",
            params={},
            data=None,
        )
        obj = LicenseStatus(**response.json())
        obj.auth = self.auth
        return obj
